import typing as tp
from collections import abc
from itertools import chain
from sys import getsizeof
from typing import NamedTuple

import numpy as np

from static_frame.core.display_config import DisplayConfig
from static_frame.core.util import DTYPE_OBJECT_KIND
from static_frame.core.util import EMPTY_ARRAY
from static_frame.core.util import bytes_to_size_label

if tp.TYPE_CHECKING:
    from static_frame.core.frame import Frame  # pylint: disable=W0611 #pragma: no cover


class MFConfig(NamedTuple):
    local_only: bool # only data locally owned by arrays, or all referenced data
    materialized: bool # measure byte payload nbytes (regardless of sharing)
    data_only: bool # only array byte payloads, or all objects


class MeasureFormatMeta(type):
    def __iter__(cls) -> tp.Iterator[MFConfig]:
        return (v for (k, v) in vars(cls).items() if not k.startswith('_'))

class MeasureFormat(metaclass=MeasureFormatMeta):
    LOCAL = MFConfig(
            local_only=True,
            materialized=False,
            data_only=False,
            )
    LOCAL_MATERIALIZED = MFConfig(
            local_only=True,
            materialized=True,
            data_only=False,
            )
    LOCAL_MATERIALIZED_DATA = MFConfig(
            local_only=True,
            materialized=True,
            data_only=True,
            )
    # NOTE MFConfig(local_only=True, materialized=False, data_only=True) is invalid

    REFERENCED = MFConfig(
            local_only=False,
            materialized=False,
            data_only=False,
            )
    REFERENCED_MATERIALIZED = MFConfig(
            local_only=False,
            materialized=True,
            data_only=False,
            )
    REFERENCED_MATERIALIZED_DATA = MFConfig(
            local_only=False,
            materialized=True,
            data_only=True,
            )
    # NOTE MFConfig(local_only=False, materialized=False, data_only=True) is invalid

FORMAT_TO_DISPLAY = {
        MeasureFormat.LOCAL: 'L',
        MeasureFormat.LOCAL_MATERIALIZED: 'LM',
        MeasureFormat.LOCAL_MATERIALIZED_DATA: 'LMD',
        MeasureFormat.REFERENCED: 'R',
        MeasureFormat.REFERENCED_MATERIALIZED: 'RM',
        MeasureFormat.REFERENCED_MATERIALIZED_DATA: 'RMD',
        }

class MaterializedArray:
    '''Wrapper of array that delivers the sizeof as the fully realized size, ignoring any potential sharing of memory.
    '''

    __slots__ = (
            '_array',
            '_format',
            )
    BASE_ARRAY_BYTES = getsizeof(EMPTY_ARRAY)

    def __init__(self,
            array: np.ndarray,
            format: MFConfig = MeasureFormat.LOCAL,
            ):
        self._array = array
        self._format = format

    def __sizeof__(self) -> int:
        size = 0
        if self._format.local_only and self._array.base is not None:
            pass # all data referenced externally
        else:
            size += self._array.nbytes

        if not self._format.data_only:
            size += self.BASE_ARRAY_BYTES

        return size


class MemoryMeasure:

    @staticmethod
    def _iter_iterable(obj: tp.Any) -> tp.Iterator[tp.Any]:
        '''
        Generates the iterable children that have not been counted by a getsizeof call on the parent object
        '''
        if hasattr(obj, '__iter__') and not isinstance(obj, str):
            if obj.__class__ is np.ndarray and obj.dtype.kind == DTYPE_OBJECT_KIND:
                # NOTE: iter(obj) would return slices for multi-dimensional arrays
                yield from (obj[loc] for loc in np.ndindex(obj.shape))
            elif isinstance(obj, (abc.Sequence, abc.Set)):
                yield from obj
            elif isinstance(obj, dict):
                yield from chain.from_iterable(obj.items())
            else:
                # The full size of the object is included in its getsizeof call
                # e.g. SF Containers, FrozenAutoMap, integer numpy arrays, int, float, etc.
                pass

    @staticmethod
    def _iter_slots(obj: tp.Any) -> tp.Iterator[tp.Any]:
        '''
        Generates an iterable of the values of all slot-based attributes in an object, including the slots contained in the object's parent classes based on the MRO
        '''
        # NOTE: This does NOT support 'single-string' slots (i.e. __slots__ = 'foo')
        slots = chain.from_iterable(
                cls.__slots__ for cls in obj.__class__.__mro__
                if hasattr(cls, '__slots__'))
        attrs = (getattr(obj, slot) for slot in slots
                if slot != '__weakref__' and hasattr(obj, slot))
        yield from attrs

    @classmethod
    def nested_sizable_elements(cls,
            obj: tp.Any,
            *,
            format: MFConfig = MeasureFormat.REFERENCED,
            seen: tp.Set[int],
            ) -> tp.Iterator[tp.Any]:
        '''
        Generates an iterable of all objects the parent object has references to, including nested references. This function considers both the iterable unsized children (based on _iter_iterable) and the sizable
        attributes listed in its slots. The resulting generator is in pre-order and includes the parent object at the end.
        '''
        from static_frame.core.container import ContainerBase

        if id(obj) in seen:
            return
        seen.add(id(obj))

        if obj.__class__ is np.ndarray:
            if format.materialized:
                obj = MaterializedArray(obj, format=format)
            else: # non-object arrays report included elements
                if obj.dtype.kind == DTYPE_OBJECT_KIND:
                    for el in cls._iter_iterable(obj):
                        yield from cls.nested_sizable_elements(el, seen=seen, format=format)
                if not format.local_only and obj.base is not None:
                    # include the base array for numpy slices / views only if that base has not been seen
                    yield from cls.nested_sizable_elements(obj.base, seen=seen, format=format)
            yield obj

        elif obj.__class__ is MaterializedArray:
            yield obj

        else:
            for el in cls._iter_iterable(obj): # will not yield if no __iter__
                yield from cls.nested_sizable_elements(el, seen=seen, format=format)
            # arrays do not have slots
            for el in cls._iter_slots(obj):
                yield from cls.nested_sizable_elements(el, seen=seen, format=format)
            yield obj


def memory_total(
        obj: tp.Any,
        *,
        format: MFConfig = MeasureFormat.REFERENCED,
        seen: tp.Union[None, tp.Set[tp.Any]] = None,
        ) -> int:
    '''
    Returns the total size of the object and its references, including nested refrences
    '''
    seen = set() if seen is None else seen

    def gen() -> tp.Iterator[int]:
        for component in MemoryMeasure.nested_sizable_elements(obj,
                seen=seen,
                format=format,
                ):
            if format.data_only and component.__class__ is MaterializedArray:
                yield component.__sizeof__() # call directly to avoid gc ovehead addition
            else:
                yield getsizeof(component)

    return sum(gen())

class MemoryDisplay:
    '''A simple container for capturing and displaying memory usage in bytes for StaticFrame containers.
    '''

    __slots__ = (
            '_frame',
            '_repr',
            )

    @classmethod
    def from_any(cls,
            obj: tp.Any,
            label_component_pairs: tp.Iterable[tp.Tuple[str, tp.Any]] = (),
            ) -> 'MemoryDisplay':
        '''Given any slotted object, return a :obj:`MemoryDisplay` instance.

        '''
        from static_frame.core.frame import Frame

        parts = chain(label_component_pairs, (('Total', obj),))

        def gen() -> tp.Iterator[tp.Tuple[tp.Tuple[str, ...], tp.List[int]]]:
            for label, part in parts:
                sizes = []
                for format in MeasureFormat:
                    # NOTE: not sharing seen accross evaluations
                    sizes.append(memory_total(part, format=format))
                yield label, sizes

        if hasattr(obj, 'name') and obj.name is not None:
            name = f'<{obj.__class__.__name__}: {obj.name}>'
        else:
            name = f'<{obj.__class__.__name__}>'

        f = Frame.from_records_items(
                gen(),
                columns=(FORMAT_TO_DISPLAY[mf] for mf in MeasureFormat),
                name=name,
                )
        return cls(f)

    def __init__(self, frame: 'Frame'):
        '''Initialize an instance with a ``Frame`` of byte counts.
        '''
        from static_frame.core.frame import Frame
        self._frame = frame

        f_size = self._frame.iter_element().apply(bytes_to_size_label)

        def gen() -> tp.Iterator[tp.Sequence[str]]:
            for row_old in f_size.iter_tuple(axis=1):
                row_new = []
                for e in row_old:
                    row_new.extend(e.split(' '))
                yield row_new

        f = Frame.from_records(gen(), index=f_size.index)
        columns = [
                f_size.columns[i//2] if i % 2 == 0
                else f'{f_size.columns[i//2]}u'.ljust(5)
                for i in f.columns
                ]
        f = f.relabel(columns=columns)
        dc = DisplayConfig(type_show=False)
        self._repr: str = f.display(config=dc).__repr__()

    def __repr__(self) -> str:
        return self._repr

    def to_frame(self) -> 'Frame':
        '''Return a Frame of byte counts.
        '''
        return self._frame




import frame_fixtures as ff
import numpy as np

from static_frame.core.index import Index
from static_frame.core.protocol_dfi import ArrowCType
from static_frame.core.protocol_dfi import DFIBuffer
from static_frame.core.protocol_dfi import DFIColumn
from static_frame.core.protocol_dfi import DFIDataFrame
from static_frame.core.protocol_dfi import np_dtype_to_dfi_dtype
from static_frame.core.protocol_dfi_abc import ColumnNullType
from static_frame.core.protocol_dfi_abc import DlpackDeviceType
from static_frame.core.protocol_dfi_abc import DtypeKind
from static_frame.core.util import NAT
from static_frame.test.test_case import TestCase


class TestUnit(TestCase):

    def test_arrow_ctype_a(self) -> None:
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.float64)), 'g')
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.float32)), 'f')
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.float16)), 'e')

        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.int64)), 'l')
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.int8)), 'c')

        self.assertEqual(ArrowCType.from_dtype(np.dtype(bool)), 'b')

        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.uint64)), 'L')
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.uint8)), 'C')

    def test_arrow_ctype_b(self) -> None:
        with self.assertRaises(NotImplementedError):
            ArrowCType.from_dtype(np.dtype(object))

    def test_arrow_ctype_c(self) -> None:
        self.assertEqual(ArrowCType.from_dtype(np.dtype(str)), 'u')

    def test_arrow_ctype_d(self) -> None:
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.datetime64('2022-01-01'))), 'tdm')

    def test_arrow_ctype_e(self) -> None:
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.datetime64('2022-01-01T01:01:01'))), 'tts')
        self.assertEqual(ArrowCType.from_dtype(np.dtype(np.datetime64('2022-01-01', 'ns'))), 'ttn')

    def test_arrow_ctype_f(self) -> None:
        with self.assertRaises(NotImplementedError):
            ArrowCType.from_dtype(np.dtype(np.datetime64('2022-01')))

    def test_arrow_ctype_g(self) -> None:
        with self.assertRaises(NotImplementedError):
            ArrowCType.from_dtype(np.dtype(complex))

    #---------------------------------------------------------------------------
    def test_np_dtype_to_dfi_dtype_a(self) -> None:
        self.assertEqual(np_dtype_to_dfi_dtype(
                np.dtype(bool)),
                (DtypeKind.BOOL, 8, 'b', '='),
                )

    def test_np_dtype_to_dfi_dtype_b(self) -> None:
        self.assertEqual(np_dtype_to_dfi_dtype(
                np.dtype(np.float64)),
                (DtypeKind.FLOAT, 64, 'g', '='),
                )

    def test_np_dtype_to_dfi_dtype_c(self) -> None:
        self.assertEqual(np_dtype_to_dfi_dtype(
                np.dtype(np.uint8)),
                (DtypeKind.UINT, 8, 'C', '='),
                )

    #---------------------------------------------------------------------------
    def test_dfi_buffer_a(self) -> None:
        dfib = DFIBuffer(np.array((True, False)))
        self.assertEqual(str(dfib), '<DFIBuffer: shape=(2,) dtype=|b1>')
        self.assertTrue(dfib.__array__().data.contiguous)

    def test_dfi_buffer_b(self) -> None:
        # only accept already-contiguous arrays
        with self.assertRaises(ValueError):
            dfib = DFIBuffer((np.arange(12).reshape(6, 2) % 3 == 0)[:, 0])

    def test_dfi_buffer_array_a(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        self.assertEqual(dfib.__array__().tolist(), a1.tolist())

    def test_dfi_buffer_array_b(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        self.assertEqual(dfib.__array__(str).tolist(), a1.astype(str).tolist())

    def test_dfi_buffer_bufsize_a(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        self.assertEqual(dfib.bufsize, 2)

    def test_dfi_buffer_ptr_a(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        self.assertEqual(dfib.ptr, a1.__array_interface__['data'][0])

    def test_dfi_buffer_dlpack_a(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        with self.assertRaises(NotImplementedError):
            dfib.__dlpack__()

    def test_dfi_buffer_dlpack_device_a(self) -> None:
        a1 = np.array((True, False))
        dfib = DFIBuffer(a1)
        self.assertEqual(dfib.__dlpack_device__(), (DlpackDeviceType.CPU, None))

    #---------------------------------------------------------------------------

    def test_dfi_column_init_a(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(str(dfic), '<DFIColumn: shape=(2,) dtype=|b1>')

    def test_dfi_column_array_a(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.__array__().tolist(), a1.tolist())

    def test_dfi_column_array_b(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.__array__(str).tolist(), a1.astype(str).tolist())

    def test_dfi_column_size_a(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.size(), 2)

    def test_dfi_column_offset_a(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.offset, 0)

    def test_dfi_column_dtype_a(self) -> None:
        a1 = np.array((True, False))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.dtype, (DtypeKind.BOOL, 8, 'b', '='))

    def test_dfi_column_dtype_b(self) -> None:
        a1 = np.array((1.1, 2.2), dtype=np.float64)
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.dtype, (DtypeKind.FLOAT, 64, 'g', '='))

    def test_dfi_column_dtype_c(self) -> None:
        a1 = np.array((1.1, 2.2), dtype=np.float16)
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.dtype, (DtypeKind.FLOAT, 16, 'e', '='))

    def test_dfi_column_describe_categorical_a(self) -> None:
        a1 = np.array((1.1, 2.2), dtype=np.float64)
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        with self.assertRaises(TypeError):
            dfic.describe_categorical()

    def test_dfi_column_describe_null_a(self) -> None:
        a1 = np.array((1.1, 2.2, np.nan), dtype=np.float64)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.describe_null, (ColumnNullType.USE_NAN, None))

    def test_dfi_column_describe_null_b(self) -> None:
        a1 = np.array(('2020-01-01', '2022-05-01', NAT), dtype=np.datetime64)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.describe_null, (ColumnNullType.USE_SENTINEL, NAT))
        post = dfic.get_buffers()

    def test_dfi_column_describe_null_c(self) -> None:
        a1 = np.array((3, 4))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.describe_null, (ColumnNullType.NON_NULLABLE, None))

    def test_dfi_column_null_count_a(self) -> None:
        a1 = np.array((1.1, 2.2, np.nan), dtype=np.float64)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.null_count, 1)

    def test_dfi_column_null_count_b(self) -> None:
        a1 = np.array(('2020-01', '2022-05', NAT), dtype=np.datetime64)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.null_count, 1)

    def test_dfi_column_null_count_c(self) -> None:
        a1 = np.array((3, 4))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.null_count, 0)

    def test_dfi_column_metadata_a(self) -> None:
        a1 = np.array((3, 4))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        [(mk, mv)] = dfic.metadata.items()
        self.assertEqual(mk, 'static-frame.index')
        self.assertTrue(mv.equals(mv))

    def test_dfi_column_num_chunks_a(self) -> None:
        a1 = np.array((3, 4))
        idx1 = Index(('a', 'b'))
        dfic = DFIColumn(a1, idx1)
        self.assertEqual(dfic.num_chunks(), 1)

    def test_dfi_column_chunks_a(self) -> None:
        a1 = np.arange(5)
        idx1 = Index(('a', 'b', 'c', 'd', 'e'))
        dfic = DFIColumn(a1, idx1)
        post = tuple(dfic.get_chunks(2))

        self.assertEqual(
                [c.__array__().tolist() for c in post],
                [[0, 1, 2], [3, 4]],
                )

    def test_dfi_column_chunks_b(self) -> None:
        a1 = np.arange(5)
        idx1 = Index(('a', 'b', 'c', 'd', 'e'))
        dfic = DFIColumn(a1, idx1)
        post = tuple(dfic.get_chunks(5))

        self.assertEqual(
                [c.__array__().tolist() for c in post],
                [[0], [1], [2], [3], [4]],
                )

    def test_dfi_column_chunks_c(self) -> None:
        a1 = np.arange(5)
        idx1 = Index(('a', 'b', 'c', 'd', 'e'))
        dfic = DFIColumn(a1, idx1)
        post = tuple(dfic.get_chunks(1))

        self.assertEqual(
                [c.__array__().tolist() for c in post],
                [[0, 1, 2, 3, 4]],
                )

    def test_dfi_column_get_buffers_a(self) -> None:
        a1 = np.array((1.1, 2.2, np.nan), dtype=np.float64)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        post = dfic.get_buffers()

        assert post['data'] is not None

        self.assertEqual(str(post['data'][0]), '<DFIBuffer: shape=(3,) dtype=<f8>')
        self.assertEqual(post['data'][1], (DtypeKind.FLOAT, 64, 'g', '='))

        assert post['validity'] is not None

        self.assertEqual(str(post['validity'][0]), '<DFIBuffer: shape=(3,) dtype=|b1>')
        self.assertEqual(post['validity'][1], (DtypeKind.BOOL, 8, 'b', '='))

        self.assertEqual(post['offsets'], None)

    def test_dfi_column_get_buffers_b(self) -> None:
        a1 = np.array((False, True, False), dtype=bool)
        idx1 = Index(('a', 'b', 'c'))
        dfic = DFIColumn(a1, idx1)
        post = dfic.get_buffers()

        assert post['data'] is not None

        self.assertEqual(str(post['data'][0]), '<DFIBuffer: shape=(3,) dtype=|b1>')
        self.assertEqual(post['data'][1], (DtypeKind.BOOL, 8, 'b', '='))

        self.assertEqual(post['validity'], None)
        self.assertEqual(post['offsets'], None)

    #---------------------------------------------------------------------------

    def test_dfi_df_init_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)')
        dfif1 = DFIDataFrame(f)
        dfif2 = dfif1.__dataframe__()
        self.assertEqual(str(dfif1), str(dfif2))

    def test_dfi_df_array_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool)')
        dfif = DFIDataFrame(f)
        post = dfif.__array__(str).tolist()
        self.assertEqual(post,
                [['False', 'False'], ['False', 'False'], ['False', 'False']])

    def test_dfi_df_metadata_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)').rename('foo')
        dfif = DFIDataFrame(f)
        [(mk, mv), (mnk, mnv)] = dfif.metadata.items()
        self.assertEqual(mk, 'static-frame.index')
        self.assertTrue(mv.equals(f.index))
        self.assertEqual(mnk, 'static-frame.name')
        self.assertEqual(mnv, f.name)

    def test_dfi_df_num_columns_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)')
        dfif = DFIDataFrame(f)
        self.assertEqual(dfif.num_columns(), 2)

    def test_dfi_df_num_rows_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)')
        dfif = DFIDataFrame(f)
        self.assertEqual(dfif.num_rows(), 3)

    def test_dfi_df_num_chunks_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)')
        dfif = DFIDataFrame(f)
        self.assertEqual(dfif.num_chunks(), 1)

    def test_dfi_df_column_names_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)')
        dfif = DFIDataFrame(f)
        self.assertEqual(tuple(dfif.column_names()), (0, 1))

    def test_dfi_df_get_column_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)|c(I,str)')
        dfif = DFIDataFrame(f)
        self.assertEqual(str(dfif.get_column(0)), '<DFIColumn: shape=(3,) dtype=|b1>')

    def test_dfi_df_get_column_by_name_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)|c(I,str)')
        dfif = DFIDataFrame(f)
        self.assertEqual(str(dfif.get_column_by_name('zZbu')), '<DFIColumn: shape=(3,) dtype=|b1>')

    def test_dfi_df_get_columns_a(self) -> None:
        f = ff.parse('s(3,2)|v(bool,float)|c(I,str)')
        dfif = DFIDataFrame(f)
        self.assertEqual(
                [str(c) for c in dfif.get_columns()],
                ['<DFIColumn: shape=(3,) dtype=|b1>', '<DFIColumn: shape=(3,) dtype=<f8>']
                )

    def test_dfi_df_select_columns_a(self) -> None:
        f = ff.parse('s(3,4)|v(bool,float)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        dfif2 = dfif1.select_columns([0, 3])
        self.assertEqual(str(dfif2), '<DFIDataFrame: shape=(3, 2)>')

    def test_dfi_df_select_columns_b(self) -> None:
        f = ff.parse('s(3,4)|v(bool,float)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        dfif2 = dfif1.select_columns(range(2))
        self.assertEqual(str(dfif2), '<DFIDataFrame: shape=(3, 2)>')


    def test_dfi_df_select_columns_by_name_a(self) -> None:
        f = ff.parse('s(3,4)|v(bool,float)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        dfif2 = dfif1.select_columns_by_name(['ztsv', 'zkuW'])
        self.assertEqual(str(dfif2), '<DFIDataFrame: shape=(3, 2)>')
        self.assertEqual(tuple(dfif2.column_names()), ('ztsv', 'zkuW'))

    def test_dfi_df_select_columns_by_name_b(self) -> None:
        f = ff.parse('s(3,4)|v(bool,float)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        dfif2 = dfif1.select_columns_by_name((x for x in ('ztsv', 'zkuW')))
        self.assertEqual(str(dfif2), '<DFIDataFrame: shape=(3, 2)>')
        self.assertEqual(tuple(dfif2.column_names()), ('ztsv', 'zkuW'))


    def test_dfi_df_get_chunks_a(self) -> None:
        f = ff.parse('s(5,4)|v(bool,int64)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        post = [df.__array__().tolist() for df in dfif1.get_chunks(2)]
        self.assertEqual(post,
            [[[False, 162197, True, 129017], [False, -41157, False, 35021], [False, 5729, False, 166924]], [[True, -168387, True, 122246], [False, 140627, False, 197228]]])

    def test_dfi_df_get_chunks_b(self) -> None:
        f = ff.parse('s(5,4)|v(bool,int64)|c(I,str)')
        dfif1 = DFIDataFrame(f)
        post = [str(df) for df in dfif1.get_chunks(1)]
        self.assertEqual(post, ['<DFIDataFrame: shape=(5, 4)>'])




if __name__ == '__main__':
    import unittest
    unittest.main()





import typing as tp
from datetime import date
from datetime import datetime

import numpy as np
from arraykit import isna_element
from arraykit import resolve_dtype

from static_frame.core.node_selector import Interface
from static_frame.core.node_selector import InterfaceBatch
from static_frame.core.node_selector import TContainer
from static_frame.core.util import DT64_AS
from static_frame.core.util import DT64_DAY
from static_frame.core.util import DT64_FS
from static_frame.core.util import DT64_H
from static_frame.core.util import DT64_M
from static_frame.core.util import DT64_MONTH
from static_frame.core.util import DT64_MS
from static_frame.core.util import DT64_NS
from static_frame.core.util import DT64_PS
from static_frame.core.util import DT64_S
from static_frame.core.util import DT64_US
from static_frame.core.util import DT64_YEAR
from static_frame.core.util import DTYPE_DATETIME_KIND
from static_frame.core.util import DTYPE_INT_DEFAULT
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_STR
from static_frame.core.util import DTYPE_STR_KINDS
from static_frame.core.util import DTYPE_YEAR_MONTH_STR
from static_frame.core.util import FILL_VALUE_DEFAULT
from static_frame.core.util import AnyCallable
from static_frame.core.util import array_from_element_apply
from static_frame.core.util import array_from_element_attr
from static_frame.core.util import array_from_element_method
from static_frame.core.util import dtype_from_element
from static_frame.core.util import isna_array

if tp.TYPE_CHECKING:
    from static_frame.core.batch import Batch  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import Frame  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index import Index  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.series import Series  # pylint: disable = W0611 #pragma: no cover
    from static_frame.core.type_blocks import TypeBlocks  # pylint: disable = W0611 #pragma: no cover

BlocksType = tp.Iterable[np.ndarray]
ToContainerType = tp.Callable[[tp.Iterator[np.ndarray]], TContainer]

INTERFACE_DT = (
        '__call__',
        'year',
        'year_month',
        'month',
        'day',
        'hour',
        'minute',
        'second',
        'weekday',
        'quarter',
        'is_month_end',
        'is_month_start',
        'is_year_end',
        'is_year_start',
        'is_quarter_end',
        'is_quarter_start',
        'timetuple',
        'isoformat',
        'fromisoformat',
        'strftime',
        'strptime',
        'strpdate',
        )

class InterfaceDatetime(Interface[TContainer]):

    __slots__ = (
            '_blocks', # function that returns iterable of arrays
            '_blocks_to_container', # partialed function that will return a new container
            '_fill_value',
            '_fill_value_dtype',
            )
    INTERFACE = INTERFACE_DT

    DT64_EXCLUDE_YEAR = (DT64_YEAR,)
    DT64_EXCLUDE_YEAR_MONTH = (DT64_YEAR, DT64_MONTH)
    DT64_TIME = frozenset((
            DT64_H,
            DT64_M,
            DT64_S,
            DT64_MS,
            DT64_US,
            DT64_NS,
            DT64_PS,
            DT64_FS,
            DT64_AS,
            ))

    DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO = frozenset((
            DT64_YEAR,
            DT64_MONTH,
            DT64_NS,
            DT64_PS,
            DT64_FS,
            DT64_AS,
            ))

    def __init__(self,
            *,
            blocks: BlocksType,
            blocks_to_container: ToContainerType[TContainer],
            fill_value: tp.Any = FILL_VALUE_DEFAULT,
            ) -> None:
        self._blocks: BlocksType = blocks
        self._blocks_to_container: ToContainerType[TContainer] = blocks_to_container
        self._fill_value: tp.Any = fill_value
        self._fill_value_dtype: tp.Optional[np.dtype] = (None
                if fill_value is FILL_VALUE_DEFAULT
                else dtype_from_element(fill_value))

    def __call__(self,
            *,
            fill_value: tp.Any,
            ) -> 'InterfaceDatetime[TContainer]':
        '''
        Args:
            fill_value: If NAT are encountered, use this value.
        '''
        return self.__class__(
                blocks=self._blocks,
                blocks_to_container=self._blocks_to_container,
                fill_value=fill_value,
                )

    @staticmethod
    def _validate_dtype_non_str(
            dtype: np.dtype,
            exclude: tp.Iterable[np.dtype] = (),
            ) -> None:
        '''
        Only support dtypes that are (or contain) datetime64 types. This is because most conversions from string can be done simply with astype().
        '''
        if ((dtype.kind == DTYPE_DATETIME_KIND
                or dtype == DTYPE_OBJECT)
                and dtype not in exclude
                ):
            return
        raise RuntimeError(f'invalid dtype ({dtype}) for date operation')

    @staticmethod
    def _validate_dtype_str(
            dtype: np.dtype,
            exclude: tp.Iterable[np.dtype] = (),
            ) -> None:
        '''
        Only support dtypes that are (or contain) strings.
        '''
        if ((dtype.kind in DTYPE_STR_KINDS
                or dtype == DTYPE_OBJECT)
                and dtype not in exclude
                ):
            return
        raise RuntimeError(f'invalid dtype ({dtype}) for operation on string types')

    def _fill_missing_dt64(self, array_src: np.ndarray, array_dst: np.ndarray) -> np.ndarray:
        '''
        Args:
            array_src: The raw array, before any dytpe conversions; used to identify missing values.
            array_dst: The array post any conversions, to be filled with missing values.
        '''
        if self._fill_value is not FILL_VALUE_DEFAULT:
            targets = isna_array(array_src)
            if targets.any():
                dt = resolve_dtype(array_dst.dtype, self._fill_value_dtype)
                if dt != array_dst.dtype:
                    array_dst = array_dst.astype(dt)
                array_dst[targets] = self._fill_value

        array_dst.flags.writeable = False
        return array_dst

    def _fill_missing_element_method(self,
            array: np.ndarray,
            *,
            method_name: str,
            args: tp.Tuple[tp.Any, ...],
            dtype: np.dtype,
            ) -> np.ndarray:
        if self._fill_value is FILL_VALUE_DEFAULT:
            array = array_from_element_method(
                    array=array,
                    method_name=method_name,
                    args=args,
                    dtype=dtype,
                    )
        else:
            dt = resolve_dtype(dtype, self._fill_value_dtype)
            if dtype.itemsize == 0 and dt.kind == dtype.kind:
                dt = dtype # set to unsized

            def func(e: tp.Any) -> tp.Any:
                if isna_element(e):
                    return self._fill_value
                return getattr(e, method_name)(*args)

            array = array_from_element_apply(
                    array=array,
                    func=func,
                    dtype=dt,
                    )
        assert array.flags.writeable is False
        return array

    def _fill_missing_element_attr(self,
            array: np.ndarray,
            *,
            attr_name: str,
            dtype: np.dtype,
            ) -> np.ndarray:
        if self._fill_value is FILL_VALUE_DEFAULT:
            array = array_from_element_attr(
                    array=array,
                    attr_name=attr_name,
                    dtype=dtype,
                    )
        else:
            dt = resolve_dtype(dtype, self._fill_value_dtype)
            assert dtype.itemsize != 0 # expect numeric scalar

            def func(e: tp.Any) -> tp.Any:
                if isna_element(e):
                    return self._fill_value
                return getattr(e, attr_name)

            array = array_from_element_apply(
                    array=array,
                    func=func,
                    dtype=dt,
                    )
        assert array.flags.writeable is False
        return array

    #---------------------------------------------------------------------------
    # date, datetime attributes

    @property
    def year(self) -> TContainer:
        'Return the year of each element.'

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    array = block.astype(DT64_YEAR).astype(DTYPE_INT_DEFAULT) + 1970
                    array = self._fill_missing_dt64(block, array)
                else: # must be object type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='year',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def month(self) -> TContainer:
        '''
        Return the month of each element, between 1 and 12 inclusive.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    array = block.astype(DT64_MONTH).astype(DTYPE_INT_DEFAULT) % 12 + 1
                    array = self._fill_missing_dt64(block, array)
                else: # must be object type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='month',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def year_month(self) -> TContainer:
        '''
        Return the year and month of each element as string formatted YYYY-MM.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    array = block.astype(DT64_MONTH).astype(DTYPE_YEAR_MONTH_STR)
                    array = self._fill_missing_dt64(block, array)
                else:
                    array = self._fill_missing_element_method(block,
                            method_name='strftime',
                            args=('%Y-%m',),
                            dtype=DTYPE_YEAR_MONTH_STR,
                            )
                yield array

        return self._blocks_to_container(blocks())


    @property
    def day(self) -> TContainer:
        '''
        Return the day of each element, between 1 and the number of days in the given month of the given year.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_DAY:
                        block = block.astype(DT64_DAY)
                    # subtract the first of the month, then shift
                    array = (block - block.astype(DT64_MONTH)).astype(DTYPE_INT_DEFAULT) + 1
                    array = self._fill_missing_dt64(block, array)
                else: # must be object type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='day',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    #---------------------------------------------------------------------------
    # datetime attributes

    @property
    def hour(self) -> TContainer:
        '''
        Return the hour of each element, between 0 and 24.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit all dt64 types
                self._validate_dtype_non_str(block.dtype)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_H:
                        block = block.astype(DT64_H)
                    # subtract the first of the month, then shfit
                    array = block.astype(DTYPE_INT_DEFAULT) % 24
                    array = self._fill_missing_dt64(block, array)
                else: # must be object datetime type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='hour',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def minute(self) -> TContainer:
        '''
        Return the minute of each element, between 0 and 60.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit all dt64 types
                self._validate_dtype_non_str(block.dtype)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_M:
                        block = block.astype(DT64_M)
                    array = block.astype(DTYPE_INT_DEFAULT) % 60
                    array = self._fill_missing_dt64(block, array)
                else: # must be object datetime type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='minute',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def second(self) -> TContainer:
        '''
        Return the second of each element, between 0 and 60.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit all dt64 types
                self._validate_dtype_non_str(block.dtype)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_S:
                        block = block.astype(DT64_S)
                    # subtract the first of the month, then shfit
                    array = block.astype(DTYPE_INT_DEFAULT) % 60
                    array = self._fill_missing_dt64(block, array)
                else: # must be object datetime type
                    array = self._fill_missing_element_attr(
                            array=block,
                            attr_name='second',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())


    #---------------------------------------------------------------------------

    # replace: awkward to implement, as cannot provide None for the parameters that you do not want to set

    def weekday(self) -> TContainer:
        '''
        Return the day of the week as an integer, where Monday is 0 and Sunday is 6.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_DAY: # go to day first, then object
                        block = block.astype(DT64_DAY)
                    # shift to set first Monday, then modulo
                    array = (block.astype(DTYPE_INT_DEFAULT) + 3) % 7
                    array = self._fill_missing_dt64(block, array)
                else:
                    # NOTE: might be faster to convert to datetime64 then do shift / modulo
                    array = self._fill_missing_element_method(
                            array=block,
                            method_name='weekday',
                            args=(),
                            dtype=DTYPE_INT_DEFAULT
                            )
                yield array

        return self._blocks_to_container(blocks())

    def quarter(self) -> TContainer:
        '''
        Return the quarter of the year as an integer, where January through March is quarter 1.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype)
                # astype object dtypes to month too
                if block.dtype != DT64_MONTH:
                    b = block.astype(DT64_MONTH)
                else:
                    b = block
                # months will start from 0
                b = b.astype(DTYPE_INT_DEFAULT) % 12
                is_q1 = b <= 2 # 0-2
                is_q4 = b >= 9 # 9-11
                is_q2 = (b <= 5) & ~is_q1 #3-5

                array = np.full(block.shape, 3, dtype=DTYPE_INT_DEFAULT)
                array[is_q1] = 1
                array[is_q4] = 4
                array[is_q2] = 2

                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    #---------------------------------------------------------------------------
    # boolean matches

    def is_month_end(self) -> TContainer:
        '''Return Boolean indicators if the day is the month end.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block
                # convert to month, shift to next, convert to day, slide back to eom
                array = b == ((b.astype(DT64_MONTH) + 1).astype(DT64_DAY) - 1)
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    def is_month_start(self) -> TContainer:
        '''Return Boolean indicators if the day is the month start.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block
                array = b == b.astype(DT64_MONTH).astype(DT64_DAY)
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())


    def is_year_end(self) -> TContainer:
        '''Return Boolean indicators if the day is the year end.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block
                # convert to year, shift to next, convert to day, slide back to eoy
                array = b == ((b.astype(DT64_YEAR) + 1).astype(DT64_DAY) - 1)
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    def is_year_start(self) -> TContainer:
        '''Return Boolean indicators if the day is the year start.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block
                # convert to month, shift to next, convert to day, slide back to eom
                array = b == b.astype(DT64_YEAR).astype(DT64_DAY)
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())


    def is_quarter_end(self) -> TContainer:
        '''Return Boolean indicators if the day is the quarter end.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block

                # convert to month, shift to next, convert to day, slide back to eom
                month = b.astype(DT64_MONTH)
                eom = (month + 1).astype(DT64_DAY) - 1
                # months starting from 0
                month_int = month.astype(DTYPE_INT_DEFAULT) % 12
                month_valid = ((month_int == 2)
                        | (month_int == 5)
                        | (month_int == 8)
                        | (month_int == 11)
                        )
                array = (b == eom) & month_valid
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    def is_quarter_start(self) -> TContainer:
        '''Return Boolean indicators if the day is the quarter start.
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                # astype object dtypes to day too
                if block.dtype != DT64_DAY:
                    b = block.astype(DT64_DAY)
                else:
                    b = block

                # convert to month, shift to next, convert to day, slide back to eom
                month = b.astype(DT64_MONTH)
                som = month.astype(DT64_DAY)
                # months starting from 0
                month_int = month.astype(DTYPE_INT_DEFAULT) % 12
                month_valid = ((month_int == 0)
                        | (month_int == 3)
                        | (month_int == 6)
                        | (month_int == 9)
                        )
                array = (b == som) & month_valid
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    #---------------------------------------------------------------------------
    # time methods

    def timetuple(self) -> TContainer:
        '''
        Return a ``time.struct_time`` such as returned by time.localtime().
        '''
        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                # NOTE: nanosecond and lower will return integers; should exclude
                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    block = block.astype(DTYPE_OBJECT)
                # all object arrays by this point
                array = self._fill_missing_element_method(
                        array=block,
                        method_name='timetuple',
                        args=(),
                        dtype=DTYPE_OBJECT
                        )
                yield array

        return self._blocks_to_container(blocks())

    def isoformat(self, sep: str = 'T', timespec: str = 'auto') -> TContainer:
        '''
        Return a string representing the date in ISO 8601 format, YYYY-MM-DD.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                args = ()
                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype in self.DT64_TIME:
                        # if we know this is a time type, we can pass args
                        args = (sep, timespec) #type: ignore
                    block = block.astype(DTYPE_OBJECT)

                # all object arrays by this point
                # NOTE: we cannot determine if an Object array has date or datetime objects with a full iteration, so we cannot be sure if we need to pass args or not.
                array = self._fill_missing_element_method(
                        array=block,
                        method_name='isoformat',
                        args=args,
                        dtype=DTYPE_STR,
                        )
                yield array

        return self._blocks_to_container(blocks())

    def fromisoformat(self) -> TContainer:
        '''
        Return a :obj:`datetime.date` object from an ISO 8601 format.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                self._validate_dtype_str(block.dtype)

                # NOTE: might use fromisoformat on date/datetime objects directly; assumed to be faster to go through datetime64 objects, fromisoformat is only available on python 3.7

                array_dt64 = block.astype(np.datetime64)
                if array_dt64.dtype in self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO:
                    raise RuntimeError(f'invalid derived dtype ({array_dt64.dtype}) for iso format')
                array = array_dt64.astype(object)
                array = self._fill_missing_dt64(block, array)
                yield array

        return self._blocks_to_container(blocks())

    def strftime(self, format: str) -> TContainer:
        '''
        Return a string representing the date, controlled by an explicit ``format`` string.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                # NOTE: nanosecond and lower will return integers; should exclud
                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    block = block.astype(DTYPE_OBJECT)
                # all object arrays by this point

                # returns an immutable array
                array = self._fill_missing_element_method(
                        array=block,
                        method_name='strftime',
                        args=(format,),
                        dtype=DTYPE_STR,
                        )
                yield array

        return self._blocks_to_container(blocks())

    def strptime(self, format: str) -> TContainer:
        '''
        Return a Python datetime object from parsing a string defined with ``format``.
        '''
        def func(s: str) -> datetime:
            return datetime.strptime(s, format)

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                # NOTE: no missing handling necessary
                self._validate_dtype_str(block.dtype)
                # returns an immutable array
                array = array_from_element_apply(
                        array=block,
                        func=func,
                        dtype=DTYPE_OBJECT,
                        )
                yield array

        return self._blocks_to_container(blocks())


    def strpdate(self, format: str) -> TContainer:
        '''
        Return a Python date object from parsing a string defined with ``format``.
        '''
        def func(s: str) -> date:
            return datetime.strptime(s, format).date()

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                self._validate_dtype_str(block.dtype)
                # returns an immutable array
                array = array_from_element_apply(
                        array=block,
                        func=func,
                        dtype=DTYPE_OBJECT,
                        )
                yield array

        return self._blocks_to_container(blocks())

#-------------------------------------------------------------------------------

class InterfaceBatchDatetime(InterfaceBatch):
    '''Alternate datetime interface specialized for the :obj:`Batch`.
    '''
    __slots__ = (
            '_batch_apply',
            '_fill_value',
            )
    INTERFACE = INTERFACE_DT

    def __init__(self,
            batch_apply: tp.Callable[[AnyCallable], 'Batch'],
            *,
            fill_value: tp.Any = FILL_VALUE_DEFAULT,
            ) -> None:
        self._batch_apply = batch_apply
        self._fill_value = fill_value


    def __call__(self,
            *,
            fill_value: tp.Any,
            ) -> 'InterfaceBatchDatetime':
        '''
        Args:
            fill_value: If NAT are encountered, use this value.
        '''
        return self.__class__(self._batch_apply,
                fill_value=fill_value
                )

    #---------------------------------------------------------------------------
    # date, datetime attributes

    @property
    def year(self) -> 'Batch':
        'Return the year of each element.'
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).year)

    @property
    def month(self) -> 'Batch':
        '''
        Return the month of each element, between 1 and 12 inclusive.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).month)

    @property
    def year_month(self) -> 'Batch':
        '''
        Return the month of each element, between 1 and 12 inclusive.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).year_month)

    @property
    def day(self) -> 'Batch':
        '''
        Return the day of each element, between 1 and the number of days in the given month of the given year.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).day)

    #---------------------------------------------------------------------------
    # datetime attributes

    @property
    def hour(self) -> 'Batch':
        '''
        Return the hour of each element, between 0 and 24.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).hour)

    @property
    def minute(self) -> 'Batch':
        '''
        Return the minute of each element, between 0 and 60.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).minute)

    @property
    def second(self) -> 'Batch':
        '''
        Return the second of each element, between 0 and 60.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).second)

    #---------------------------------------------------------------------------

    # replace: akward to implement, as cannot provide None for the parameters that you do not want to set

    def weekday(self) -> 'Batch':
        '''
        Return the day of the week as an integer, where Monday is 0 and Sunday is 6.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).weekday())

    def quarter(self) -> 'Batch':
        '''
        Return the quarter of the year as an integer, where January through March is quarter 1.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).quarter())

    #---------------------------------------------------------------------------
    # boolean matches

    def is_month_end(self) -> 'Batch':
        '''Return Boolean indicators if the day is the month end.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_month_end())

    def is_month_start(self) -> 'Batch':
        '''Return Boolean indicators if the day is the month start.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_month_start())

    def is_year_end(self) -> 'Batch':
        '''Return Boolean indicators if the day is the year end.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_year_end())

    def is_year_start(self) -> 'Batch':
        '''Return Boolean indicators if the day is the year start.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_year_start())

    def is_quarter_end(self) -> 'Batch':
        '''Return Boolean indicators if the day is the quarter end.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_quarter_end())

    def is_quarter_start(self) -> 'Batch':
        '''Return Boolean indicators if the day is the quarter start.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).is_quarter_start())

    #---------------------------------------------------------------------------
    # time methods

    def timetuple(self) -> 'Batch':
        '''
        Return a ``time.struct_time`` such as returned by time.localtime().
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).timetuple())

    def isoformat(self, sep: str = 'T', timespec: str = 'auto') -> 'Batch':
        '''
        Return a string representing the date in ISO 8601 format, YYYY-MM-DD.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).isoformat(sep, timespec))

    def fromisoformat(self) -> 'Batch':
        '''
        Return a :obj:`datetime.date` object from an ISO 8601 format.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).fromisoformat())

    def strftime(self, format: str) -> 'Batch':
        '''
        Return a string representing the date, controlled by an explicit ``format`` string.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).strftime(format))

    def strptime(self, format: str) -> 'Batch':
        '''
        Return a Python datetime object from parsing a string defined with ``format``.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).strptime(format))

    def strpdate(self, format: str) -> 'Batch':
        '''
        Return a Python date object from parsing a string defined with ``format``.
        '''
        return self._batch_apply(lambda c: c.via_dt(fill_value=self._fill_value).strpdate(format))



















import unittest
from sys import getsizeof

import frame_fixtures as ff
import numpy as np
from automap import FrozenAutoMap  # pylint: disable=E0611

from static_frame.core.memory_measure import MaterializedArray
from static_frame.core.memory_measure import MeasureFormat
from static_frame.core.memory_measure import MemoryDisplay
from static_frame.core.memory_measure import MemoryMeasure
from static_frame.core.memory_measure import memory_total
from static_frame.test.test_case import TestCase

_iter_iterable = MemoryMeasure._iter_iterable
_iter_slots = MemoryMeasure._iter_slots
nested_sizable_elements = MemoryMeasure.nested_sizable_elements

class TestUnit(TestCase):

    #---------------------------------------------------------------------------
    # MemoryMeasure._iter_iterable

    def test_unsized_children_none(self) -> None:
        self.assertEqual(tuple(_iter_iterable(None)), ())

    def test_unsized_children_int(self) -> None:
        self.assertEqual(tuple(_iter_iterable(3)), ())

    def test_unsized_children_str(self) -> None:
        self.assertEqual(tuple(_iter_iterable('abc')), ())

    def test_unsized_children_float(self) -> None:
        self.assertEqual(tuple(_iter_iterable(4.5)), ())

    def test_unsized_children_numpy_array_int(self) -> None:
        obj = np.array([2, 3, 4], dtype=np.int64)
        self.assertEqual(tuple(_iter_iterable(obj)), ())

    def test_unsized_children_numpy_array_object(self) -> None:
        obj = np.array([2, 'a', 4], dtype=object)
        self.assertEqual(tuple(_iter_iterable(obj)), (2, 'a', 4))

    def test_unsized_children_tuple(self) -> None:
        obj = (2, 3, 4)
        self.assertEqual(tuple(_iter_iterable(obj)), (2, 3, 4))

    def test_unsized_children_tuple_nested(self) -> None:
        obj = (2, ('a', 'b', ('c', 'd')), 4)
        self.assertEqual(tuple(_iter_iterable(obj)), (2, ('a', 'b', ('c', 'd')), 4))

    def test_unsized_children_list(self) -> None:
        obj = [2, 3, 4]
        self.assertEqual(tuple(_iter_iterable(obj)), (2, 3, 4))

    def test_unsized_children_list_nested(self) -> None:
        obj = [2, ('a', 'b', ('c', 'd')), 4]
        self.assertEqual(tuple(_iter_iterable(obj)), (2, ('a', 'b', ('c', 'd')), 4))

    def test_unsized_children_set(self) -> None:
        obj = set((2, 3, 4))
        self.assertEqual(tuple(_iter_iterable(obj)), (2, 3, 4))

    def test_unsized_children_frozenset(self) -> None:
        obj = frozenset((2, 3, 4))
        self.assertEqual(tuple(_iter_iterable(obj)), (2, 3, 4))

    def test_unsized_children_dict(self) -> None:
        obj = { 'a': 2, 'b': 3, 'c': 4 }
        self.assertEqual(tuple(_iter_iterable(obj)), ('a', 2, 'b', 3, 'c', 4))

    def test_unsized_children_frozenautomap(self) -> None:
        obj = FrozenAutoMap([2, 3, 4])
        self.assertEqual(tuple(_iter_iterable(obj)), ())

    def test_unsized_children_numpy_array_object_complex_has_unique_ids(self) -> None:
        # make sure that all elements are looped through in a multi-dimensional object array
        obj = np.array([np.array([None, None, i]) for i in range(10)])
        self.assertEqual(len(set(id(el) for el in _iter_iterable(obj))), 11)

    #---------------------------------------------------------------------------
    # MemoryMeasure._iter_slots

    def test_sizable_slot_attrs_empty(self) -> None:
        class A:
            pass
        obj = A()
        self.assertEqual(tuple(_iter_slots(obj)), ())

    def test_sizable_slot_attrs_simple(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
                'carrots',
                'dumplings',
                'eggs'
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
                self.carrots = 'c'
                self.dumplings = 'd'
                self.eggs = 'e'
        obj = A()
        self.assertEqual(frozenset(_iter_slots(obj)), frozenset(('a', 'b', 'c', 'd', 'e')))

    def test_sizable_slot_attrs_not_all_initialized(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
                'carrots',
                'dumplings',
                'eggs'
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
                self.eggs = 'e'
        obj = A()
        self.assertEqual(frozenset(_iter_slots(obj)), frozenset(('a', 'b', 'e')))

    def test_sizable_slot_attrs_inheritance_1_layer(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
        class B(A):
            __slots__ = (
                'carrots',
                'dumplings',
                'eggs'
            )
            def __init__(self) -> None:
                super().__init__()
                self.carrots = 'c'
                self.dumplings = 'd'
                self.eggs = 'e'
        obj = B()
        self.assertEqual(frozenset(_iter_slots(obj)), frozenset(('a', 'b', 'c', 'd', 'e')))

    def test_sizable_slot_attrs_inheritance_1_layer_overlapping_slots(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
                'carrots',
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
        class B(A):
            __slots__ = ( # pylint: disable=W0244 # intentionally redefining 'carrots'
                'dumplings',
                'eggs'
            )
            def __init__(self) -> None:
                super().__init__()
                self.carrots = 'c'
                self.dumplings = 'd'
                self.eggs = 'e'
        obj = B()
        self.assertEqual(frozenset(_iter_slots(obj)), frozenset(('a', 'b', 'c', 'd', 'e')))
        self.assertEqual(len(tuple(_iter_slots(obj))), 5)

    def test_sizable_slot_attrs_inheritance_2_layers(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
        class B(A):
            __slots__ = (
                'carrots',
                'dumplings',
            )
            def __init__(self) -> None:
                super().__init__()
                self.carrots = 'c'
                self.dumplings = 'd'
        class C(B):
            __slots__ = (
                'eggs',
            )
            def __init__(self) -> None:
                super().__init__()
                self.eggs = 'e'
        obj = C()
        sizables = frozenset(_iter_slots(obj))
        self.assertEqual(sizables, frozenset(('a', 'b', 'c', 'd', 'e')))

    def test_sizable_slot_attrs_inheritance_multiple(self) -> None:
        class A:
            __slots__ = (
                'apples',
                'bananas',
                'carrots',
                'dumplings',
            )
            def __init__(self) -> None:
                self.apples = 'a'
                self.bananas = 'b'
                self.carrots = 'c'
                self.dumplings = 'd'
        class B:
            __slots__ = ()
        class C(A, B):
            __slots__ = (
                'eggs',
            )
            def __init__(self) -> None:
                super().__init__()
                self.eggs = 'e'
        obj = C()
        self.assertEqual(frozenset(_iter_slots(obj)), frozenset(('a', 'b', 'c', 'd', 'e')))

    # NOTE: From https://docs.python.org/3/reference/datamodel.html#notes-on-using-slots
    # > Multiple inheritance with multiple slotted parent classes can be used, but only one
    # >   parent is allowed to have attributes created by slots (the other bases must have empty
    # >   slot layouts) - violations raise TypeError.

    #---------------------------------------------------------------------------
    # MemoryMeasure.nested_sizable_elements

    def test_nested_sizable_elements_none(self) -> None:
        self.assertEqual(tuple(nested_sizable_elements(None, seen=set())), (None,))

    def test_nested_sizable_elements_int(self) -> None:
        self.assertEqual(tuple(nested_sizable_elements(3, seen=set())), (3,))

    def test_nested_sizable_elements_str(self) -> None:
        self.assertEqual(tuple(nested_sizable_elements('abc', seen=set())), ('abc',))

    def test_nested_sizable_elements_float(self) -> None:
        self.assertEqual(tuple(nested_sizable_elements(4.5, seen=set())), (4.5,))

    def test_nested_sizable_elements_numpy_array_int(self) -> None:
        obj = np.array([2, 3, 4], dtype=np.int64)
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (obj,))

    def test_nested_sizable_elements_numpy_array_object(self) -> None:
        obj = np.array([2, 'a', 4], dtype=object)
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (2, 'a', 4, obj))

    def test_nested_sizable_elements_tuple(self) -> None:
        obj = (2, 3, 4)
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (2, 3, 4, obj))

    def test_nested_sizable_elements_tuple_nested(self) -> None:
        obj = (2, ('a', 'b', ('c', 'd')), 4)
        self.assertEqual(
            tuple(nested_sizable_elements(obj, seen=set())),
            (2, 'a', 'b', 'c', 'd', ('c', 'd'), ('a', 'b', ('c', 'd')), 4, obj)
        )

    def test_nested_sizable_elements_tuple_nested_existing_seen(self) -> None:
        cd = ('c', 'd')
        obj = (2, ('a', 'b', cd), 4)
        seen = set(id(el) for el in ('a', 'c', 'd', cd))
        self.assertEqual(
            tuple(nested_sizable_elements(obj, seen=seen)),
            (2, 'b', ('a', 'b', ('c', 'd')), 4, obj)
        )

    def test_nested_sizable_elements_list(self) -> None:
        obj = [2, 3, 4]
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (2, 3, 4, obj))

    def test_nested_sizable_elements_list_nested(self) -> None:
        obj = [2, ('a', 'b', ('c', 'd')), 4]
        self.assertEqual(
            tuple(nested_sizable_elements(obj, seen=set())),
            (2, 'a', 'b', 'c', 'd', ('c', 'd'), ('a', 'b', ('c', 'd')), 4, obj)
        )

    def test_nested_sizable_elements_set(self) -> None:
        obj = set((2, 3, 4))
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (2, 3, 4, obj))

    def test_nested_sizable_elements_frozenset(self) -> None:
        obj = frozenset((2, 3, 4))
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (2, 3, 4, obj))

    def test_nested_sizable_elements_dict(self) -> None:
        obj = { 'a': 2, 'b': 3, 'c': 4 }
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), ('a', 2, 'b', 3, 'c', 4, obj))

    def test_nested_sizable_elements_frozenautomap(self) -> None:
        obj = FrozenAutoMap([2, 3, 4])
        self.assertEqual(tuple(nested_sizable_elements(obj, seen=set())), (obj,))

    #---------------------------------------------------------------------------
    def test_measure_format_a(self) -> None:
        empty = np.array(())
        a1 = np.array((1, 2), dtype=np.int64)
        a2 = a1[:]

        mempty = MaterializedArray(empty, format=MeasureFormat.LOCAL_MATERIALIZED_DATA)
        ma1 = MaterializedArray(a1, format=MeasureFormat.LOCAL_MATERIALIZED_DATA)

        self.assertEqual(memory_total(mempty,
                format=MeasureFormat.REFERENCED_MATERIALIZED_DATA),
                0)
        self.assertEqual(memory_total(empty,
                format=MeasureFormat.REFERENCED_MATERIALIZED_DATA),
                0)

        self.assertEqual(memory_total(ma1,
                format=MeasureFormat.REFERENCED_MATERIALIZED_DATA),
                a1.nbytes,
                )
        self.assertEqual(memory_total(a1,
                format=MeasureFormat.REFERENCED_MATERIALIZED_DATA),
                a1.nbytes,
                )
        self.assertEqual(memory_total(a2,
                format=MeasureFormat.REFERENCED_MATERIALIZED_DATA),
                a1.nbytes,
                )

    def test_measure_format_b(self) -> None:
        empty = np.array(())
        mempty = MaterializedArray(empty, format=MeasureFormat.REFERENCED_MATERIALIZED)

        a1 = np.array((1, 2), dtype=np.int64)
        ma1 = MaterializedArray(a1, format=MeasureFormat.REFERENCED_MATERIALIZED)

        a2 = a1[:]
        ma2 = MaterializedArray(a2, format=MeasureFormat.REFERENCED_MATERIALIZED)

        # import ipdb; ipdb.set_trace()
        self.assertEqual(memory_total(empty,
                format=MeasureFormat.REFERENCED_MATERIALIZED),
                getsizeof(mempty),
                )

        self.assertEqual(memory_total(a1,
                format=MeasureFormat.REFERENCED_MATERIALIZED),
                getsizeof(ma1)
                )

        self.assertEqual(memory_total(a2,
                format=MeasureFormat.REFERENCED_MATERIALIZED),
                getsizeof(ma2)
                )

    def test_measure_format_c(self) -> None:
        empty = np.array(())
        a1 = np.array((1, 2), dtype=np.int64)
        a2 = a1[:]

        self.assertEqual(memory_total(empty,
                format=MeasureFormat.LOCAL),
                getsizeof(empty),
                )

        self.assertEqual(memory_total(a1,
                format=MeasureFormat.LOCAL),
                getsizeof(a1),
                )

        self.assertEqual(memory_total(a2,
                format=MeasureFormat.LOCAL),
                getsizeof(a2),
                )

    def test_measure_format_d(self) -> None:
        empty = np.array(())
        a1 = np.array((1, 2), dtype=np.int64)
        a2 = a1[:]

        self.assertEqual(memory_total(empty,
                format=MeasureFormat.REFERENCED),
                getsizeof(empty),
                )

        self.assertEqual(memory_total(a1,
                format=MeasureFormat.REFERENCED),
                getsizeof(a1),
                )

        self.assertEqual(memory_total(a2,
                format=MeasureFormat.REFERENCED),
                getsizeof(a2) + getsizeof(a1),
                )


    def test_measure_format_e(self) -> None:
        empty = np.array(())
        a1 = np.array((1, 2), dtype=np.int64)
        a2 = a1[:]

        self.assertEqual(memory_total(empty,
                format=MeasureFormat.LOCAL_MATERIALIZED_DATA),
                0,
                )

        self.assertEqual(memory_total(a1,
                format=MeasureFormat.LOCAL_MATERIALIZED_DATA),
                a1.nbytes,
                )

        self.assertEqual(memory_total(a2,
                format=MeasureFormat.LOCAL_MATERIALIZED_DATA),
                0,
                )

    def test_measure_format_f(self) -> None:
        empty = np.array(())
        a1 = np.array((1, 2), dtype=np.int64)
        a2 = a1[:]

        mempty = MaterializedArray(empty, format=MeasureFormat.LOCAL_MATERIALIZED)
        ma1 = MaterializedArray(a1, format=MeasureFormat.LOCAL_MATERIALIZED)
        ma2 = MaterializedArray(a2, format=MeasureFormat.LOCAL_MATERIALIZED)

        self.assertEqual(memory_total(empty,
                format=MeasureFormat.LOCAL_MATERIALIZED),
                getsizeof(mempty),
                )

        self.assertEqual(memory_total(a1,
                format=MeasureFormat.LOCAL_MATERIALIZED),
                getsizeof(ma1),
                )

        self.assertEqual(memory_total(a2,
                format=MeasureFormat.LOCAL_MATERIALIZED),
                getsizeof(ma2),
                )

    #---------------------------------------------------------------------------


    def test_memory_display_a(self) -> None:
        f = ff.parse('s(16,8)|i(I,str)|v(str,int,float)')

        post = MemoryDisplay.from_any(f,
                (('Index', f._index), ('Columns', f._columns), ('Values', f._blocks)),
                )
        self.assertEqual(post.to_frame().loc['Total']['R'], memory_total(f, format=MeasureFormat.REFERENCED))

if __name__ == '__main__':
    unittest.main()



import typing as tp
import unittest
from sys import getsizeof

import frame_fixtures as ff
import numpy as np
from automap import FrozenAutoMap  # pylint: disable=E0611

from static_frame import Bus
from static_frame import Frame
from static_frame import Index
from static_frame import IndexDateGO
from static_frame import IndexGO
from static_frame import IndexHierarchy
from static_frame import Quilt
from static_frame import Series
from static_frame import StoreConfig
from static_frame import TypeBlocks
from static_frame import Yarn
from static_frame.core.memory_measure import memory_total
from static_frame.test.test_case import TestCase
from static_frame.test.test_case import temp_file


class TestUnit(TestCase):
    def test_getsizeof_total_int(self) -> None:
        obj = 2
        self.assertEqual(memory_total(obj), getsizeof(2))

    def test_getsizeof_total_set(self) -> None:
        obj = set(['a', 'b', 4])
        self.assertEqual(
            memory_total(obj),
            sum(getsizeof(e) for e in (
                'a', 'b', 4,
                set(['a', 'b', 4]),
            ))
        )

    def test_getsizeof_total_frozenset(self) -> None:
        obj = frozenset(['a', 'b', 4])
        self.assertEqual(
            memory_total(obj),
            sum(getsizeof(e) for e in (
                'a', 'b', 4,
                frozenset(['a', 'b', 4]),
            ))
        )

    def test_getsizeof_total_np_array(self) -> None:
        obj = np.arange(3)
        self.assertEqual(
            memory_total(obj),
            getsizeof(obj)
        )

    def test_getsizeof_total_np_array_slice(self) -> None:
        arr = np.arange(5)
        obj = arr[2:]
        self.assertEqual(
            memory_total(obj),
            sum(getsizeof(e) for e in (
                obj,
                obj.base,
            ))
        )

    def test_getsizeof_total_np_array_with_slice(self) -> None:
        arr = np.arange(5)
        sli = arr[2:]
        obj = [arr, sli]
        self.assertEqual(
            memory_total(obj),
            sum(getsizeof(e) for e in (
                arr,
                sli,
                obj,
            ))
        )

    def test_getsizeof_total_frozenautomap(self) -> None:
        obj = FrozenAutoMap(['a', 'b', 'c'])
        self.assertEqual(
            memory_total(obj),
            getsizeof(FrozenAutoMap(['a', 'b', 'c']))
        )

    def test_getsizeof_total_dict(self) -> None:
        obj = { 'a': 2, 'b': 3, 'c': (4, 5) }
        self.assertEqual(
            memory_total(obj),
            sum(getsizeof(e) for e in (
                'a', 2,
                'b', 3,
                'c', 4, 5, (4, 5),
                { 'a': 2, 'b': 3, 'c': (4, 5) },
            ))
        )

    def test_getsizeof_total_tuple(self) -> None:
        obj = (2, 3, 4)
        self.assertEqual(memory_total(obj), sum(getsizeof(e) for e in (
            2, 3, 4,
            (2, 3, 4),
        )))

    def test_getsizeof_total_nested_tuple(self) -> None:
        obj = (2, 'b', (2, 3))
        self.assertEqual(memory_total(obj), sum(getsizeof(e) for e in (
            2, 'b', 3,
            (2, 3),
            (2, 'b', (2, 3)),
        )))

    def test_getsizeof_total_predefined_seen(self) -> None:
        obj = (4, 5, (2, 8))
        seen = set((id(2), id(3)))
        self.assertEqual(memory_total(obj, seen=seen), sum(getsizeof(e) for e in (
            4, 5,
            8,
            (2, 8),
            (4, 5, (2, 8))
        )))

    def test_getsizeof_total_predefined_seen_all_elements(self) -> None:
        tup_a = (2, 8)
        obj = (4, 5, tup_a)
        seen = set((id(e) for e in (4, 5, 2, 8, tup_a, obj)))
        self.assertEqual(memory_total(obj, seen=seen), 0)

    def test_getsizeof_total_predefined_seen_sub_elements(self) -> None:
        tup_a = (2, 8)
        obj = (4, 5, tup_a)
        seen = set((id(e) for e in (4, 5, 2, 8, tup_a)))
        self.assertEqual(memory_total(obj, seen=seen), getsizeof(obj))

    def test_getsizeof_total_predefined_seen_base_element(self) -> None:
        tup_a = (2, 8)
        obj = (4, 5, tup_a)
        seen = set((id(obj),))
        self.assertEqual(memory_total(obj, seen=seen), 0)

    def test_getsizeof_total_larger_values_is_larger(self) -> None:
        a = ('ab', 'cd', 'ef')
        b = ('abc', 'def', 'ghi')
        self.assertLess(memory_total(a), memory_total(b))

    def test_getsizeof_total_more_values_is_larger_a(self) -> None:
        a = ('a', 'b', 'c')
        b = ('a', 'b', 'c', 'd')
        self.assertLess(memory_total(a), memory_total(b))

    def test_getsizeof_total_more_values_is_larger_b(self) -> None:
        a = ('a', 'b', 'c')
        b = 'd'
        self.assertLess(memory_total([a]), memory_total([a, b]))

    def test_getsizeof_total_more_values_is_larger_nested_a(self) -> None:
        a = ('a', (2, (8, 9), 4), 'c')
        b = ('a', (2, (8, 9, 10), 4), 'c')
        self.assertLess(memory_total(a), memory_total(b))

    def test_getsizeof_total_more_values_is_larger_nested_b(self) -> None:
        a = np.array(['a', [2, (8, 9), 4], 'c'], dtype=object)
        b = np.array(['a', [2, (8, 9, 10), 4], 'c'], dtype=object)
        self.assertLess(memory_total(a), memory_total(b))

    #---------------------------------------------------------------------------
    # Frame

    def test_getsizeof_total_frame_simple(self) -> None:
        f = ff.parse('s(3,4)')
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            getsizeof(f)
        )))

    def test_getsizeof_total_frame_string_index(self) -> None:
        f = ff.parse('s(3,4)|i(I,str)|c(I,str)')
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            getsizeof(f)
        )))

    def test_getsizeof_total_frame_object_index(self) -> None:
        f = ff.parse('s(8,12)|i(I,object)|c(I,str)')
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            getsizeof(f)
        )))

    def test_getsizeof_total_frame_multiple_value_types(self) -> None:
        f = ff.parse('s(8,4)|i(I,object)|c(I,str)|v(object,int,bool,str)')
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            getsizeof(f)
        )))

    def test_getsizeof_total_frame_he_before_hash(self) -> None:
        f = ff.parse('s(3,4)').to_frame_he()
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            # memory_total(f._hash, seen=seen), # not initialized yet
            getsizeof(f)
        )))

    def test_getsizeof_total_frame_he_after_hash(self) -> None:
        f = ff.parse('s(3,4)').to_frame_he()
        hash(f) # to initialize _hash
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(f), sum((
            memory_total(f._blocks, seen=seen),
            memory_total(f._columns, seen=seen),
            memory_total(f._index, seen=seen),
            memory_total(f._name, seen=seen),
            memory_total(f._hash, seen=seen),
            getsizeof(f)
        )))

    #---------------------------------------------------------------------------
    # Index

    def test_getsizeof_total_index_simple(self) -> None:
        idx = Index(('a', 'b', 'c'))
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            idx
        )))

    def test_getsizeof_total_index_object(self) -> None:
        idx = Index((1, 'b', (2, 3)))
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            # _positions is an object dtype numpy array
            1, 'b',
            2, 3,
            (2, 3),
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            idx
        )))

    def test_getsizeof_total_index_loc_is_iloc(self) -> None:
        idx = Index((0, 1, 2), loc_is_iloc=True)
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            # idx._name, # both _map and _name are None
            idx
        )))

    def test_getsizeof_total_index_empty(self) -> None:
        idx = Index(())
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            idx
        )))

    def test_getsizeof_total_index_name_adds_size(self) -> None:
        idx1 = Index(('a', 'b', 'c'))
        idx2 = idx1.rename('with_name')
        self.assertLess(memory_total(idx1), memory_total(idx2))

    def test_getsizeof_total_index_more_values_adds_size(self) -> None:
        idx1 = Index(('a', 'b', 'c'))
        idx2 = Index(('a', 'b', 'c', 'd'))
        self.assertLess(memory_total(idx1), memory_total(idx2))

    def test_getsizeof_total_index_more_nested_values_adds_size(self) -> None:
        idx1 = Index((1, 'b', (2, 3)))
        idx2 = Index((1, 'b', (2, 3, 4, 5)))
        self.assertLess(memory_total(idx1), memory_total(idx2))

    def test_getsizeof_total_index_more_doubly_nested_values_adds_size(self) -> None:
        idx1 = Index((1, 'b', ('c', (8, 9), 'd')))
        idx2 = Index((1, 'b', ('c', (8, 9, 10), 'd')))
        self.assertLess(memory_total(idx1), memory_total(idx2))

    def test_getsizeof_total_index_loc_is_iloc_reduces_size(self) -> None:
        # idx1 will be smaller since the _positions and _labels variables point to the same array
        idx1 = Index((0, 1, 2), loc_is_iloc=True)
        idx2 = Index((0, 1, 2))
        self.assertLess(memory_total(idx1), memory_total(idx2))

    def test_getsizeof_total_index_go(self) -> None:
        idx = IndexGO(('a', 'b', 'c'))
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            'a', 'b', 'c',
            idx._labels_mutable,
            idx._labels_mutable_dtype,
            idx._positions_mutable_count,
            idx
        )))

    def test_getsizeof_total_index_go_after_append(self) -> None:
        idx = IndexGO(('a', 'b', 'c'))
        idx.append('d')
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            'a', 'b', 'c', 'd',
            idx._labels_mutable,
            idx._labels_mutable_dtype,
            idx._positions_mutable_count,
            idx
        )))

    def test_getsizeof_total_index_datetime_go(self) -> None:
        idx = IndexDateGO.from_date_range('1994-01-01', '1995-01-01')
        self.assertEqual(memory_total(idx), sum(getsizeof(e) for e in (
            idx._map,
            idx._labels,
            idx._positions,
            idx._positions.base, # from PositionsAllocator.get
            idx._recache,
            idx._name,
            *idx._labels_mutable, # Note: _labels_mutable is not nested
            idx._labels_mutable,
            idx._labels_mutable_dtype,
            idx._positions_mutable_count,
            idx
        )))

    #---------------------------------------------------------------------------
    # Series

    def test_getsizeof_total_series_simple(self) -> None:
        s = Series(('a', 'b', 'c'))
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(s), sum(memory_total(e, seen=seen) for e in (
            s.values,
            s._index,
            s._name,
        )) + getsizeof(s))

    def test_getsizeof_total_series_with_index(self) -> None:
        s = Series(('a', 'b', 'c'), index=(0, 1, 2))
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(s), sum(memory_total(e, seen=seen) for e in (
            s.values,
            s._index,
            s._name,
        )) + getsizeof(s))

    def test_getsizeof_total_series_object_values(self) -> None:
        s = Series(('a', (2, (3, 4), 5), 'c'))
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(s), sum(memory_total(e, seen=seen) for e in (
            s.values,
            s._index,
            s._name,
        )) + getsizeof(s))

    def test_getsizeof_total_series_with_name_is_larger(self) -> None:
        s1 = Series(('a', 'b', 'c'))
        s2 = s1.rename('named_series')
        self.assertLess(memory_total(s1), memory_total(s2))

    def test_getsizeof_total_series_larger_series_is_larger_a(self) -> None:
        s1 = Series(('a', 'b', 'c'))
        s2 = Series(('a', 'b', 'c', 'd'))
        self.assertLess(memory_total(s1), memory_total(s2))

    def test_getsizeof_total_series_larger_series_is_larger_b(self) -> None:
        s1 = Series(('a', 'b', 'c'))
        s2 = Series(('abc', 'def', 'ghi'))
        self.assertLess(memory_total(s1), memory_total(s2))

    def test_getsizeof_total_series_larger_nested_series_is_larger(self) -> None:
        s1 = Series(('a', (2, (4, 5), 8), 'c'))
        s2 = Series(('a', (2, (4, 5, 6), 8), 'c'))
        self.assertLess(memory_total(s1), memory_total(s2))

    def test_getsizeof_total_series_larger_index_is_larger(self) -> None:
        s1 = Series(('a', 'b', 'c'), index=(0, 1, 2))
        s2 = Series(('a', 'b', 'c'), index=('abc', 'def', 'ghi'))
        self.assertLess(memory_total(s1), memory_total(s2))

    def test_getsizeof_total_series_he_before_hash(self) -> None:
        s = Series(('a', 'b', 'c')).to_series_he()
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(s), sum(memory_total(e, seen=seen) for e in (
            s.values,
            s._index,
            s._name,
            # s._hash, # not initialized yet
        )) + getsizeof(s))

    def test_getsizeof_total_series_he_after_hash(self) -> None:
        s = Series(('a', 'b', 'c')).to_series_he()
        hash(s) # to initialize _hash
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(s), sum(memory_total(e, seen=seen) for e in (
            s.values,
            s._index,
            s._name,
            s._hash,
        )) + getsizeof(s))

    #---------------------------------------------------------------------------
    # TypeBlocks

    def test_getsizeof_total_type_blocks_1d_array(self) -> None:
        a = np.array([1, 2, 3])
        tb = TypeBlocks.from_blocks(a)
        self.assertTrue(memory_total(tb), sum(getsizeof(e) for e in (
            np.array([1, 2, 3]),
            tb._blocks, # [np.array([1, 2, 3])],
            0,
            (0, 0),
            tb._index, # [(0, 0)],
            3, 1,
            tb._shape, # (3, 1),
            np.dtype('int64'),
            tb._dtypes, # [np.dtype('int64')],
            # _row_dtype, # np.dtype('int64') is already included
            tb
        )))

    def test_getsizeof_total_type_blocks_list_of_1d_arrays(self) -> None:
        tb = TypeBlocks.from_blocks([
            np.array([1, 2, 3]),
            np.array([4, 5, 6])
        ])
        self.assertEqual(memory_total(tb), sum(getsizeof(e) for e in (
            np.array([1, 2, 3]),
            np.array([4, 5, 6]),
            tb._blocks, # [np.array([1, 2, 3]), np.array([4, 5, 6])],
            0,
            (0, 0),
            1,
            (1, 0),
            tb._index, # [(0, 0), (1, 0)],
            3, 2,
            tb._shape, # (3, 2),
            np.dtype('int64'),
            tb._dtypes, # [np.dtype('int64'), np.dtype('int64')],
            # _row_dtype, # np.dtype('int64') is already included
            tb
        )))

    def test_getsizeof_total_type_blocks_2d_array(self) -> None:
        tb = TypeBlocks.from_blocks(np.array([[1, 2, 3], [4, 5, 6]]))
        self.assertEqual(memory_total(tb), sum(getsizeof(e) for e in (
            np.array([[1, 2, 3],[4, 5, 6]]),
            tb._blocks, # [np.array([[1, 2, 3],[4, 5, 6]])],
            0,
            (0, 0),
            1,
            (0, 1),
            2,
            (0, 2),
            tb._index, # [(0, 0), (0, 1), (0, 2)],
            3,
            tb._shape, # (2, 3),
            np.dtype('int64'),
            tb._dtypes, #[np.dtype('int64'), np.dtype('int64'), np.dtype('int64')],
            # _row_dtype, # np.dtype('int64') is already included
            tb
        )))

    #---------------------------------------------------------------------------
    # IndexHierarchy

    def test_getsizeof_total_index_hierarchy_simple(self) -> None:
        idxa = Index(('a', 'b', 'c'))
        idxb = Index((1, 2, 3))
        idx = IndexHierarchy.from_product(idxa, idxb)
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(idx), sum(memory_total(e, seen=seen) for e in (
            idx._indices,
            idx._indexers,
            idx._name,
            idx._blocks,
            idx._recache,
            idx._values,
            idx._map,
            idx._index_types,
            idx._pending_extensions,
        )) + getsizeof(idx))

    #---------------------------------------------------------------------------
    # Bus

    def test_getsizeof_total_bus_simple(self) -> None:
        f1 = ff.parse('s(3,6)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        b = Bus.from_frames((f1, f2))
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(b), sum(memory_total(e, seen=seen) for e in (
            b._loaded,
            b._loaded_all,
            b._values_mutable,
            b._index,
            b._name,
            b._store,
            b._config,
            # b._last_accessed, # not initialized, not a "max_persist" bus
            b._max_persist,
        )) + getsizeof(b))

    def test_getsizeof_total_bus_maxpersist(self) -> None:
        def items() -> tp.Iterator[tp.Tuple[str, Frame]]:
            for i in range(20):
                yield str(i), Frame(np.arange(i, i+10).reshape(2, 5))

        s = Series.from_items(items(), dtype=object)
        b1 = Bus.from_series(s)

        config = StoreConfig(
                index_depth=1,
                columns_depth=1,
                include_columns=True,
                include_index=True
                )

        with temp_file('.zip') as fp:
            b1.to_zip_pickle(fp)

            b2 = Bus.from_zip_pickle(fp, config=config, max_persist=3)

            seen: tp.Set[int] = set()
            self.assertEqual(memory_total(b2), sum(memory_total(e, seen=seen) for e in (
                b2._loaded,
                b2._loaded_all,
                b2._values_mutable,
                b2._index,
                b2._name,
                b2._store,
                b2._config,
                b2._last_accessed,
                b2._max_persist,
            )) + getsizeof(b2))

    #---------------------------------------------------------------------------
    # Yarn

    def test_getsizeof_total_yarn_simple(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3))

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5))

        y = Yarn((b1, b2))
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(y), sum(memory_total(e, seen=seen) for e in (
            y._series,
            y._hierarchy,
            y._index,
            y._deepcopy_from_bus,
        )) + getsizeof(y))

    #---------------------------------------------------------------------------
    # Quilt

    def test_getsizeof_total_quilt_simple_before_columns(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,2)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,2)').rename('f4')
        f5 = ff.parse('s(4,2)').rename('f5')
        f6 = ff.parse('s(6,2)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4,), name='b2')
        b3 = Bus.from_frames((f5, f6), name='b3')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False)

        q = Quilt(y1, retain_labels=True)
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(q), sum(memory_total(e, seen=seen) for e in (
            q._bus,
            q._axis,
            q._axis_hierarchy,
            q._axis_opposite,
            q._assign_axis,
            #q._columns, # not initialized until after get columns property
            #q._index, # not initialized until after get columns property
            q._retain_labels,
            q._deepcopy_from_bus,
        )) + getsizeof(q))

    def test_getsizeof_total_quilt_simple_after_columns(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,2)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,2)').rename('f4')
        f5 = ff.parse('s(4,2)').rename('f5')
        f6 = ff.parse('s(6,2)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4,), name='b2')
        b3 = Bus.from_frames((f5, f6), name='b3')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False)

        q = Quilt(y1, retain_labels=True)
        q.columns # pylint:disable=W0104 # force columns initialization
        seen: tp.Set[int] = set()
        self.assertEqual(memory_total(q), sum(memory_total(e, seen=seen) for e in (
            q._bus,
            q._axis,
            q._axis_hierarchy,
            q._axis_opposite,
            q._assign_axis,
            q._columns,
            q._index,
            q._retain_labels,
            q._deepcopy_from_bus,
        )) + getsizeof(q))

if __name__ == '__main__':
    unittest.main()



import typing as tp

import numpy as np

from static_frame.core.protocol_dfi_abc import Buffer
from static_frame.core.protocol_dfi_abc import CategoricalDescription
from static_frame.core.protocol_dfi_abc import Column
from static_frame.core.protocol_dfi_abc import ColumnBuffers
from static_frame.core.protocol_dfi_abc import ColumnNullType
from static_frame.core.protocol_dfi_abc import DataFrame
from static_frame.core.protocol_dfi_abc import DlpackDeviceType
from static_frame.core.protocol_dfi_abc import Dtype
from static_frame.core.protocol_dfi_abc import DtypeKind
from static_frame.core.util import NAT
from static_frame.core.util import NULL_SLICE

if tp.TYPE_CHECKING:
    from static_frame import Frame  # pylint: disable=W0611 #pragma: no cover
    from static_frame import Index  # pylint: disable=W0611 #pragma: no cover


NP_KIND_TO_DFI_KIND = {
    'i': DtypeKind.INT,
    'u': DtypeKind.UINT,
    'f': DtypeKind.FLOAT,
    'b': DtypeKind.BOOL,
    'U': DtypeKind.STRING,
    'M': DtypeKind.DATETIME,
    'm': DtypeKind.DATETIME,
}

class ArrowCType:
    """
    Enum for Apache Arrow C type format strings.
    The Arrow C data interface:
    https://arrow.apache.org/docs/format/CDataInterface.html#data-type-description-format-strings
    """

    KIND_ITEMSIZE_TO_FORMAT = {
        ('b', 1): 'b',

        ('i', 1): 'c',
        ('i', 2): 's',
        ('i', 4): 'i',
        ('i', 8): 'l',

        ('u', 1): 'C',
        ('u', 2): 'S',
        ('u', 4): 'I',
        ('u', 8): 'L',

        ('f', 2): 'e',
        ('f', 4): 'f',
        ('f', 8): 'g',
    }

    NP_DT64_UNIT_TO_RESOLUTION = {
        's': 's',
        'ms': 'm',
        'us': 'u',
        'ns': 'n',
    }

    @classmethod
    def from_dtype(cls, dtype: np.dtype) -> str:
        kind = dtype.kind
        if kind == 'O':
            raise NotImplementedError('no support for object arrays')
        elif kind in ('U', 'S'):
            return 'u'
        elif kind in ('M', 'm'): # dt64
            prefix = 'tt' if kind == 'M' else 'tD' # timestamp or delta
            unit = np.datetime_data(dtype)[0]
            if unit == 'D': # a day
                return 'tdm' # NOTE: there is also tdD, date32

            try:
                res = cls.NP_DT64_UNIT_TO_RESOLUTION[unit]
            except KeyError as e:
                raise NotImplementedError(f'no support for datetime64 unit: {dtype}') from e
            return f'{prefix}{res}'

        try:
            return cls.KIND_ITEMSIZE_TO_FORMAT[(kind, dtype.itemsize)]
        except KeyError as e:
            raise NotImplementedError(f'no support for dtype: {dtype}') from e


def np_dtype_to_dfi_dtype(dtype: np.dtype) -> Dtype:
    return (NP_KIND_TO_DFI_KIND[dtype.kind],
            dtype.itemsize * 8, # bits!
            ArrowCType.from_dtype(dtype),
            '=',
            )


class DFIBuffer(Buffer):
    __slots__ = (
        '_array',
        )

    def __init__(self, array: np.ndarray) -> None:
        self._array = array

        # NOTE: would expect transformation upstream to avoid reproducing the same contiguous buffer on repeated calls;expect 1D arrays to have the same value for C_CONTIGUOUS and F_CONTIGUOUS
        if self._array.ndim != 1 or not self._array.flags['F_CONTIGUOUS']:
            raise ValueError('provided array is not contiguous')

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: shape={self._array.shape} dtype={self._array.dtype.str}>'

    def __array__(self, dtype: np.dtype = None) -> np.ndarray:
        '''
        Support the __array__ interface, returning an array of values.
        '''
        if dtype is None:
            return self._array
        return self._array.astype(dtype)

    @property
    def bufsize(self) -> int:
        return self._array.nbytes # type: ignore

    @property
    def ptr(self) -> int:
        return self._array.__array_interface__['data'][0] # type: ignore

    def __dlpack__(self) -> tp.Any:
        raise NotImplementedError("__dlpack__")

    def __dlpack_device__(self) -> tp.Tuple[DlpackDeviceType, tp.Optional[int]]:
        return (DlpackDeviceType.CPU, None)


#-------------------------------------------------------------------------------
class DFIColumn(Column):
    __slots__ = (
            '_array',
            '_index',
            )

    def __init__(self,
            array: np.ndarray,
            index: 'Index',
            ):
        assert len(array) == len(index)
        # NOTE: for efficiency, we do not store a Series, but just an array and the index (for metadata)
        self._array = array # always a 1D array
        self._index = index

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: shape={self._array.shape} dtype={self._array.dtype.str}>'

    def __array__(self, dtype: np.dtype = None) -> np.ndarray:
        '''
        Support the __array__ interface, returning an array of values.
        '''
        if dtype is None:
            return self._array
        return self._array.astype(dtype)

    def size(self) -> int:
        return self._array.size # type: ignore

    @property
    def offset(self) -> int:
        return 0

    @property
    def dtype(self) -> Dtype:
        return np_dtype_to_dfi_dtype(self._array.dtype)

    @property
    def describe_categorical(self) -> CategoricalDescription:
        raise TypeError('no categorical dtypes')

    @property
    def describe_null(self) -> tp.Tuple[ColumnNullType, tp.Any]:
        kind = self._array.dtype.kind
        if kind in ('f', 'c'):
            return (ColumnNullType.USE_NAN, None)
        elif kind in ('m', 'M'):
            return (ColumnNullType.USE_SENTINEL, NAT)
        return (ColumnNullType.NON_NULLABLE, None)

    @property
    def null_count(self) -> tp.Optional[int]:
        kind = self._array.dtype.kind
        if kind in ('f', 'c', 'm', 'M'):
            return np.isnan(self._array).sum() # type: ignore
        return 0

    @property
    def metadata(self) -> tp.Dict[str, tp.Any]:
        return {'static-frame.index': self._index}

    def num_chunks(self) -> int:
        return 1

    def get_chunks(self, n_chunks: tp.Optional[int] = None) -> tp.Iterable["DFIColumn"]:
        if n_chunks and n_chunks > 1:
            size = self._array.shape[0]
            step = size // n_chunks

            # adjust for an incomplete chunk
            if size % n_chunks != 0:
                step += 1

            for start in range(0, step * n_chunks, step):
                yield DFIColumn(
                        self._array[start: start + step],
                        self._index.iloc[start: start + step],
                        )
        else:
            yield self

    def get_buffers(self) -> ColumnBuffers:
        kind = self._array.dtype.kind
        if kind in ('f', 'c', 'm', 'M'):
            va = np.isnan(self._array)
            validity = (DFIBuffer(va), np_dtype_to_dfi_dtype(va.dtype))
        else:
            validity = None

        return dict(
                data=(DFIBuffer(self._array), self.dtype),
                validity=validity,
                offsets=None,
                )


#-------------------------------------------------------------------------------
class DFIDataFrame(DataFrame):
    __slots__ = (
            '_frame',
            '_nan_as_null',
            '_allow_copy',
            )

    def __init__(self,
            frame: 'Frame',
            *,
            nan_as_null: bool = False,
            allow_copy: bool = True,
            recast_blocks: bool = True,
            ):
        '''
        Args:
            nan_as_null: "...intended for the consumer to tell the producer to overwrite null values in the data with ``NaN``"; this is not relevant of StaticFrame does not use bit / byte masks.
            allow_copy: "... that defines whether or not the library is allowed to make a copy of the data"; this is always the case with StaticFrame as data will be made columnar contiguous.
        '''
        # NOTE: we recast internal blocks in be all contiguous columnar, meaning either 1D arrays or 2D arrays in Fortran ordering (which reduces overhead while permitting contiguous columnar slices)
        self._nan_as_null = nan_as_null
        self._allow_copy = allow_copy

        if recast_blocks:
            from static_frame.core.frame import Frame
            self._frame = Frame(
                    frame._blocks.contiguous_columnar(),
                    index=frame._index,
                    columns=frame._columns,
                    name=frame._name,
                    own_data=True,
                    own_index=True,
                    own_columns=frame.STATIC,
                    )
        else:
            assert frame.STATIC
            self._frame = frame

    def __dataframe__(self,
            nan_as_null: bool = False,
            allow_copy: bool = True,
            ) -> "DFIDataFrame":
        return self.__class__(self._frame,
                nan_as_null=self._nan_as_null,
                allow_copy=self._allow_copy,
                recast_blocks=False,
                )

    def __array__(self, dtype: np.dtype = None) -> np.ndarray:
        '''
        Support the __array__ interface, returning an array of values.
        '''
        return self._frame.__array__(dtype)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: shape={self._frame.shape}>'

    @property
    def metadata(self) -> tp.Dict[str, tp.Any]:
        return {'static-frame.index': self._frame._index,
                'static-frame.name': self._frame._name,
                }

    def num_columns(self) -> int:
        return self._frame._blocks.shape[1]

    def num_rows(self) -> tp.Optional[int]:
        return self._frame._blocks.shape[0]

    def num_chunks(self) -> int:
        return 1

    def column_names(self) -> tp.Iterable[str]:
        # NOTE: Pandas does not enforce that these are strings
        return self._frame.columns # type: ignore

    def get_column(self, i: int) -> DFIColumn:
        return DFIColumn(self._frame._blocks._extract_array_column(i), self._frame._index)

    def get_column_by_name(self, name: str) -> DFIColumn:
        return self.get_column(self._frame._columns.loc_to_iloc(name))

    def get_columns(self) -> tp.Iterable[DFIColumn]:
        index = self._frame._index
        yield from (DFIColumn(a, index) for a in self._frame._blocks.axis_values(0))

    def select_columns(self, indices: tp.Sequence[int]) -> "DFIDataFrame":
        if not isinstance(indices, list):
            indices = list(indices)

        return self.__class__(
                self._frame.iloc[NULL_SLICE, indices],
                nan_as_null=self._nan_as_null,
                allow_copy=self._allow_copy,
                recast_blocks=False,
                )

    def select_columns_by_name(self, names: tp.Sequence[str]) -> "DFIDataFrame":
        if not isinstance(names, list):
            names = list(names)

        return self.select_columns(self._frame.columns.loc_to_iloc(names))

    def get_chunks(self, n_chunks: tp.Optional[int] = None) -> tp.Iterable["DFIDataFrame"]:

        if n_chunks and n_chunks > 1:
            size = self._frame._blocks.shape[0]
            step = size // n_chunks

            # adjust for an incomplete chunk
            if size % n_chunks != 0:
                step += 1

            for start in range(0, step * n_chunks, step):
                yield DFIDataFrame(
                        self._frame.iloc[start: start + step, NULL_SLICE],
                        nan_as_null=self._nan_as_null,
                        allow_copy=self._allow_copy,
                        recast_blocks=True,
                        )
        else:
            yield self





import frame_fixtures as ff
import numpy as np

from static_frame.core.frame import Frame
from static_frame.core.index import Index
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.pivot import pivot_items_to_block
from static_frame.core.pivot import pivot_items_to_frame
from static_frame.test.test_case import TestCase

# from static_frame.core.pivot import pivot_records_items


class TestUnit(TestCase):

    def test_pivot_items_to_block_a(self) -> None:
        f = ff.parse('s(6,4)|v(int)').assign[0](
                range(6)
                )
        group_fields_iloc = [0]
        index_outer = Index(f[0].values.tolist())

        post = pivot_items_to_block(
                blocks=f._blocks,
                group_fields_iloc=group_fields_iloc,
                group_depth=1,
                data_field_iloc=3,
                func_single=None,
                dtype=np.dtype(int),
                fill_value=0,
                fill_value_dtype=np.dtype(int),
                index_outer=index_outer,
                kind='mergesort',
                )
        self.assertEqual(post.tolist(),
                [129017,  35021, 166924, 122246, 197228, 105269]
                )

    def test_pivot_items_to_frame_a(self) -> None:
        f = ff.parse('s(6,4)|v(int)').assign[0](
                range(6)
                )

        post = pivot_items_to_frame(
                blocks=f._blocks,
                group_fields_iloc=[0],
                group_depth=1,
                data_field_iloc=3,
                func_single=lambda x: str(x) if x % 2 else sum(x),
                frame_cls=Frame,
                name='foo',
                dtype=None,
                index_constructor=Index,
                columns_constructor=Index,
                kind='mergesort',
                )
        self.assertEqual(post.to_pairs(),
                (('foo', ((0, '[129017]'), (1, '[35021]'), (2, 166924), (3, 122246), (4, 197228), (5, '[105269]'))),))



    def test_pivot_items_to_frame_b(self) -> None:
        f = ff.parse('s(6,4)|v(int)').assign[0](
                range(6)
                )
        post = pivot_items_to_frame(
                blocks=f._blocks,
                group_fields_iloc=[0, 1],
                group_depth=2,
                data_field_iloc=3,
                func_single=None,
                frame_cls=Frame,
                name='foo',
                dtype=np.dtype(int),
                index_constructor=IndexHierarchy.from_labels,
                columns_constructor=Index,
                kind='mergesort',
                )
        self.assertEqual(post.to_pairs(),
                (('foo', (((0, 162197), 129017), ((1, -41157), 35021), ((2, 5729), 166924), ((3, -168387), 122246), ((4, 140627), 197228), ((5, 66269), 105269))),),
                )




    def test_pivot_core_a(self) -> None:

        frame = ff.parse('s(20,4)|v(int)').assign[0].apply(lambda s: s % 4).assign[1].apply(lambda s: s % 3)

        # by default we get a tuple index
        post1 = frame.pivot([0, 1])
        self.assertEqual(post1.index.name, (0, 1))
        self.assertIs(post1.index.__class__, Index)
        self.assertTrue(post1.to_pairs(),
                ((2, (((0, 0), 463099), ((0, 1), -88017), ((0, 2), 35021), ((1, 0), 92867), ((1, 2), 96520), ((2, 0), 172133), ((2, 1), 279191), ((2, 2), 13448), ((3, 0), 255338), ((3, 1), 372807), ((3, 2), 155574))), (3, (((0, 0), 348362), ((0, 1), 175579), ((0, 2), 105269), ((1, 0), 58768), ((1, 2), 13448), ((2, 0), 84967), ((2, 1), 239151), ((2, 2), 170440), ((3, 0), 269300), ((3, 1), 204528), ((3, 2), 493169))))
                )

        # can provide index constructor
        post2 = frame.pivot([0, 1],
                index_constructor=IndexHierarchy.from_labels)
        self.assertEqual(post2.index.name, (0, 1))
        self.assertIs(post2.index.__class__, IndexHierarchy)
        self.assertTrue(post2.to_pairs(),
                ((2, (((0, 0), 463099), ((0, 1), -88017), ((0, 2), 35021), ((1, 0), 92867), ((1, 2), 96520), ((2, 0), 172133), ((2, 1), 279191), ((2, 2), 13448), ((3, 0), 255338), ((3, 1), 372807), ((3, 2), 155574))), (3, (((0, 0), 348362), ((0, 1), 175579), ((0, 2), 105269), ((1, 0), 58768), ((1, 2), 13448), ((2, 0), 84967), ((2, 1), 239151), ((2, 2), 170440), ((3, 0), 269300), ((3, 1), 204528), ((3, 2), 493169))))
                )


if __name__ == '__main__':
    import unittest
    unittest.main()



import gzip
import os
import tempfile
import typing as tp
from io import BytesIO
from io import StringIO
from pathlib import Path
from types import TracebackType
from urllib import request
from zipfile import ZipFile

from static_frame.core.doc_str import doc_inject


class StringIOTemporaryFile(StringIO):
    '''Subclass of a StringIO that reads from a managed file that is deleted when this instance goes out of scope.
    '''

    def __init__(self, fp: Path, encoding: str) -> None:
        self._fp = fp
        self._file = open(fp, 'r', encoding=encoding) # pylint: disable=R1732
        super().__init__()

    def __del__(self) -> None:
        self._file.close()
        os.unlink(self._fp)
        super().__del__()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._file.seek(offset, whence)

    def read(self, size: tp.Optional[int] =-1) -> str:
        return self._file.read(size)

    def readline(self, size: tp.Optional[int] = -1) -> str: # type: ignore
        return self._file.readline(size)

    def __iter__(self) -> tp.Iterator[str]: # type: ignore
        return self._file.__iter__()

class BytesIOTemporaryFile(BytesIO):
    '''Subclass of a BytesIO that reads from a managed file that is deleted when this instance goes out of scope.
    '''

    def __init__(self, fp: Path) -> None:
        self._fp = fp
        self._file = open(fp, 'rb') # pylint: disable=R1732
        super().__init__()

    def __del__(self) -> None:
        self._file.close()
        os.unlink(self._fp)
        super().__del__()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._file.seek(offset, whence)

    def read(self, size: tp.Optional[int] = -1) -> bytes:
        return self._file.read(size)

    def readline(self, size: tp.Optional[int] = -1) -> bytes:
        return self._file.readline(size)

    def __iter__(self) -> tp.Iterator[bytes]:
        return self._file.__iter__()

#-------------------------------------------------------------------------------

class MaybeTemporaryFile:
    '''Provide one context manager that, if an `fp` is given, works as a normal file; if no `fp` is given, produce a temporary file.
    '''
    def __init__(self, fp: tp.Optional[Path], mode: str, encoding: str):

        if fp:
            self._f = open(fp, mode=mode, encoding=encoding) # pylint: disable=R1732
        else:
            self._f = tempfile.NamedTemporaryFile(mode=mode, # pylint: disable=R1732
                suffix=None,
                delete=False,
                encoding=encoding,
                )

    def __enter__(self) -> tp.IO[tp.Any]:
        return self._f.__enter__()

    def __exit__(self,
            type: tp.Type[BaseException],
            value: BaseException,
            traceback: TracebackType,
            ) -> None:
        self._f.__exit__(type, value, traceback)


#-------------------------------------------------------------------------------

WWWReturnType = tp.Union[Path, StringIO, BytesIO]

class WWW:
    '''Utilities for downloading resources from the world-wide-web.
    '''

    @staticmethod
    def _download_archive(
            url: tp.Union[str, request.Request],
            in_memory: bool,
            buffer_size: int,
            extension: str,
            ) -> tp.Union[Path, BytesIO]:
        archive: tp.Union[Path, BytesIO]

        with request.urlopen(url) as response:
            if in_memory:
                archive = BytesIO(response.read())
            else:
                with tempfile.NamedTemporaryFile(mode='wb',
                        suffix=extension,
                        delete=False,
                        ) as f:
                    archive = Path(f.name)
                    while True:
                        b = response.read(buffer_size)
                        if b:
                            f.write(b)
                        else:
                            break
        return archive

    @staticmethod
    def _resolve_fp_and_in_memory(
            in_memory: tp.Optional[bool],
            fp: tp.Optional[tp.Union[Path, str]] = None,
            ) -> tp.Tuple[bool, tp.Optional[Path]]:
        '''
        If an fp is given and in_memory is True, error; else, in_memory is set to False; if an fp is not given and in_memory is None, default to True, else use in_memory.
        '''
        if fp is not None:
            if in_memory is True:
                raise RuntimeError('If supplying an `fp`, `in_memory` cannot be True.')
            in_memory = False
            if isinstance(fp, str): # just to pass Path
                fp = Path(fp)
        else:
            in_memory = True if in_memory is None else in_memory
        return in_memory, fp

    @staticmethod
    def _write_maybe_temporary(
            fp: tp.Optional[Path],
            encoding: tp.Optional[str],
            extractor: tp.Callable[[], tp.Union[str, bytes]]
            ) -> WWWReturnType:
        with MaybeTemporaryFile(fp=fp,
                mode='w' if encoding else 'wb',
                encoding=encoding,
                ) as f:
            fp_written = Path(f.name)

            while True: # can use iter() function with for
                b = extractor()
                if b:
                    f.write(b)
                else:
                    break
            if fp:
                return fp_written
            if encoding:
                return StringIOTemporaryFile(fp_written, encoding=encoding)
            return BytesIOTemporaryFile(fp_written)

    #---------------------------------------------------------------------------
    @classmethod
    @doc_inject(selector='www')
    def from_file(cls,
            url: tp.Union[str, request.Request],
            *,
            encoding: tp.Optional[str] = 'utf-8',
            in_memory: tp.Optional[bool] = None,
            buffer_size: int = 8192,
            fp: tp.Optional[tp.Union[Path, str]] = None,
            ) -> WWWReturnType:
        '''
        {doc}

        Args:
            {url}
            {encoding}
            {in_memory}
            {buffer_size}
            {fp}
        '''
        in_memory, fp = cls._resolve_fp_and_in_memory(in_memory, fp)

        with request.urlopen(url) as response:
            if in_memory:
                if encoding:
                    return StringIO(response.read().decode(encoding))
                else:
                    return BytesIO(response.read())

            if encoding:
                extractor = lambda: response.read(buffer_size).decode(encoding)
            else:
                extractor = lambda: response.read(buffer_size)

            return cls._write_maybe_temporary(
                    fp=fp,
                    encoding=encoding,
                    extractor=extractor,
                    )


    @classmethod
    @doc_inject(selector='www')
    def from_zip(cls,
            url: tp.Union[str, request.Request],
            *,
            encoding: tp.Optional[str] = 'utf-8',
            in_memory: tp.Optional[bool] = None,
            buffer_size: int = 8192,
            fp: tp.Optional[tp.Union[Path, str]] = None,
            component: tp.Optional[str] = None,
            ) -> WWWReturnType:
        '''
        {doc}

        Args:
            {url}
            {encoding}
            {in_memory}
            {buffer_size}
            {fp}
            {component}
        '''
        in_memory, fp = cls._resolve_fp_and_in_memory(in_memory, fp)

        archive = cls._download_archive(url=url,
                in_memory=in_memory,
                buffer_size=buffer_size,
                extension='.zip',
                )

        with ZipFile(archive) as zf:
            names = zf.namelist()
            if component:
                name = component
            else:
                if len(names) > 1:
                    samples = ', '.join(names[:20])
                    etc = '...' if len(samples) > 20 else ''
                    raise RuntimeError(f'More than one file found in zip archive ({samples}{etc}); name a single file with the `component` argument.')
                name = names.pop()
            data = zf.read(name)

        data_io: tp.Union[StringIO, BytesIO]
        if encoding:
            data_io = StringIO(data.decode(encoding))
        else:
            data_io = BytesIO(data)

        if in_memory:
            return data_io

        # not in-memory, write a file, delete archive
        os.unlink(archive)
        return cls._write_maybe_temporary(
                fp=fp,
                encoding=encoding,
                extractor=data_io.read,
                )


    @classmethod
    @doc_inject(selector='www')
    def from_gzip(cls,
            url: tp.Union[str, request.Request],
            *,
            encoding: tp.Optional[str] = 'utf-8',
            in_memory: tp.Optional[bool] = None,
            buffer_size: int = 8192,
            fp: tp.Optional[tp.Union[Path, str]] = None,
            ) -> WWWReturnType:
        '''
        {doc}

        Args:
            {url}
            {encoding}
            {in_memory}
            {buffer_size}
            {fp}
        '''
        in_memory, fp = cls._resolve_fp_and_in_memory(in_memory, fp)

        archive = cls._download_archive(url=url,
                in_memory=in_memory,
                buffer_size=buffer_size,
                extension='.gzip',
                )

        with gzip.open(archive) as gz:
            data = gz.read()

        data_io: tp.Union[StringIO, BytesIO]
        if encoding:
            data_io = StringIO(data.decode(encoding))
        else:
            data_io = BytesIO(data)

        if in_memory:
            return data_io

        # not in-memory, write a file, delete archive
        os.unlink(archive)
        return cls._write_maybe_temporary(
                fp=fp,
                encoding=encoding,
                extractor=data_io.read,
                )




import os
import typing as tp
from functools import partial
from functools import wraps
from itertools import chain
from weakref import WeakValueDictionary

import numpy as np

from static_frame.core.exception import ErrorInitStore
from static_frame.core.exception import StoreFileMutation
from static_frame.core.exception import StoreParameterConflict
from static_frame.core.frame import Frame
from static_frame.core.store_config import StoreConfig
from static_frame.core.store_config import StoreConfigMapInitializer
from static_frame.core.util import AnyCallable
from static_frame.core.util import PathSpecifier
from static_frame.core.util import path_filter

#-------------------------------------------------------------------------------
# decorators

def store_coherent_non_write(f: AnyCallable) -> AnyCallable:

    @wraps(f)
    def wrapper(self: 'Store', *args: tp.Any, **kwargs: tp.Any) -> Frame:
        '''Decprator for derived Store class implementation of read(), labels().
        '''
        self._mtime_coherent()
        return f(self, *args, **kwargs) # type: ignore

    return wrapper


def store_coherent_write(f: AnyCallable) -> AnyCallable:
    '''Decorator for derived Store classes implementation of write()
    '''
    @wraps(f)
    def wrapper(self: 'Store', *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        post = f(self,  *args, **kwargs)
        self._mtime_update()
        return post

    return wrapper

#-------------------------------------------------------------------------------
class Store:

    _EXT: tp.FrozenSet[str]

    __slots__ = (
            '_fp',
            '_last_modified',
            '_weak_cache',
            )

    def __init__(self, fp: PathSpecifier):
        # Redefine fp variable as only string after the filter.
        fp = tp.cast(str, path_filter(fp))

        if not os.path.splitext(fp)[1] in self._EXT:
            raise ErrorInitStore(
                    f'file path {fp} does not match one of the required extensions: {self._EXT}')

        self._fp: str = fp
        self._last_modified = np.nan
        self._mtime_update()
        self._weak_cache: tp.MutableMapping[tp.Hashable, Frame] = WeakValueDictionary()

    def _mtime_update(self) -> None:
        if os.path.exists(self._fp):
            self._last_modified = os.path.getmtime(self._fp)
        else:
            self._last_modified = np.nan

    def _mtime_coherent(self) -> None:
        '''Raise if a file exists at self._fp and its mtime is not as expected
        '''
        if os.path.exists(self._fp):
            if os.path.getmtime(self._fp) != self._last_modified:
                raise StoreFileMutation(f'file {self._fp} was unexpectedly changed')
        elif not np.isnan(self._last_modified):
            # file existed previously and we got a modification time, but now it does not exist
            raise StoreFileMutation(f'expected file {self._fp} no longer exists')

    def __getstate__(self) -> tp.Tuple[None, tp.Dict[str, tp.Any]]:
        # https://docs.python.org/3/library/pickle.html#object.__getstate__
        # staying consistent with __slots__ only objects by using None as first value in tuple
        return (
            None,
            {
                attr: getattr(self, attr)
                for attr in self.__slots__
                if attr != '_weak_cache'
            }
        )

    def __setstate__(self, state: tp.Tuple[None, tp.Dict[str, tp.Any]]) -> None:
        for key, value in state[1].items():
            setattr(self, key, value)
        self._weak_cache = WeakValueDictionary()

    # def __copy__(self) -> 'Store':
    #     '''
    #     Return a new Store instance linked to the same file.
    #     '''
    #     return self.__class__(fp=self._fp)

    #---------------------------------------------------------------------------
    @staticmethod
    def get_field_names_and_dtypes(*,
            frame: Frame,
            include_index: bool,
            include_index_name: bool,
            include_columns: bool,
            include_columns_name: bool,
            force_str_names: bool = False,
            force_brackets: bool = False
            ) -> tp.Tuple[tp.Sequence[str], tp.Sequence[np.dtype]]:

        index = frame.index
        columns = frame.columns
        columns_values = columns.values

        if include_index_name and include_columns_name:
            raise StoreParameterConflict('cannot include_index_name and include_columns_name with this Store')

        if columns.depth > 1:
            # The str() of an array produces a space-delimited representation that includes list brackets; we could trim these brackets here, but need them for SQLite usage; thus, clients will have to trim if necessary.
            columns_values = tuple(str(c) for c in columns_values)

        if not include_index:
            if include_columns_name:
                raise StoreParameterConflict('cannot include_columns_name when include_index is False')
            dtypes = frame._blocks.dtypes.tolist()
            if include_columns:
                field_names = columns_values
            else: # name fields with integers?
                field_names = range(frame._blocks.shape[1])
        else:
            if index.depth == 1:
                dtypes = [index.dtype]
            else:
                dtypes = index.dtypes.values.tolist() #type: ignore [attr-defined]
            # Get a list to mutate.
            if include_index_name:
                field_names = list(index.names)
            elif include_columns_name and index.depth == columns.depth:
                field_names = list(columns.names)
            elif include_columns_name and index.depth == 1 and columns.depth > 1:
                field_names = [tuple(str(c) for c in columns.names),]
            else: # if index_depth > 1 and not equal t columns_depth
                raise StoreParameterConflict('cannot determine field names over index; set one of include_index_name or include_columns_name')

            # add frame dtypes t0 those from index
            dtypes.extend(frame._blocks.dtypes)

            # add index names in front of column names
            if include_columns:
                field_names.extend(columns_values)
            else: # name fields with integers?
                field_names.extend(range(frame._blocks.shape[1]))

        if force_str_names:
            field_names = [str(n) for n in field_names]
        if force_brackets:
            def gen() -> tp.Iterator[str]:
                for name in field_names:
                    name_str = str(name)
                    if name_str.startswith('[') and name_str.endswith(']'):
                        yield name_str
                    elif isinstance(name, tuple):
                        # make look like it came from the array
                        yield f'[{" ".join((repr(n) for n in name))}]'
                    else:
                        yield f'[{name_str}]'
            field_names = list(gen())

        return field_names, dtypes

    @staticmethod
    def _get_row_iterator(
            frame: Frame,
            include_index: bool
            ) -> tp.Callable[[], tp.Iterator[tp.Sequence[tp.Any]]]:

        if include_index:
            index = frame._index
            index_values = index.values

            def values() -> tp.Iterator[tp.Sequence[tp.Any]]:
                for idx, row in enumerate(frame.iter_array(axis=1)):
                    if index.depth > 1:
                        index_row = index_values[idx] # this is an array
                    else:
                        index_row = (index_values[idx],)
                    yield tuple(chain(index_row, row))
            return values

        return partial(frame.iter_array, axis=1) #type: ignore

    @staticmethod
    def get_column_iterator(
            frame: Frame,
            include_index: bool
            ) -> tp.Iterator[np.ndarray]:
        if include_index:
            index_depth = frame._index.depth

            if index_depth == 1:
                index_values = frame._index.values
                return chain(
                        (index_values,),
                        frame._blocks.axis_values(0)
                        )
            return chain(
                    (frame._index.values_at_depth(d) for d in range(index_depth)),
                    frame._blocks.axis_values(0)
                    )
        # avoid creating a Series per column by going to blocks
        return frame._blocks.axis_values(0)

    #---------------------------------------------------------------------------
    def read_many(self,
            labels: tp.Iterable[tp.Hashable],
            *,
            config: StoreConfigMapInitializer = None,
            container_type: tp.Type[Frame] = Frame,
            ) -> tp.Iterator[Frame]:
        '''Read many Frame, given by `labels`, from the Store. Return an iterator of instances of `container_type`.
        '''
        raise NotImplementedError() #pragma: no cover

    @store_coherent_non_write
    def read(self,
            label: tp.Hashable,
            *,
            config: tp.Optional[StoreConfig] = None,
            container_type: tp.Type[Frame] = Frame,
            ) -> Frame:
        '''Read a single Frame, given by `label`, from the Store. Return an instance of `container_type`. This is a convenience method using ``read_many``.
        '''
        return next(self.read_many((label,), config=config, container_type=container_type))

    def write(self,
            items: tp.Iterable[tp.Tuple[str, Frame]],
            *,
            config: StoreConfigMapInitializer = None
            ) -> None:
        '''Write all ``Frames`` in the Store.
        '''
        raise NotImplementedError() #pragma: no cover

    def labels(self, *,
            config: StoreConfigMapInitializer = None,
            strip_ext: bool = True,
            ) -> tp.Iterator[tp.Hashable]:
        raise NotImplementedError() #pragma: no cover

